#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: isky
@Email: 19110240019@fudan.edu.cn
@Created: 2019/10/28
------------------------------------------
@Modify: 2019/10/28
------------------------------------------
@Description: This is the basic class for pipeline component. The specific project could design
their components for specific goal, e.g., build a API KG from jar file.
"""
from abc import abstractmethod

from sekg.graph.exporter.graph_data import GraphData
from sekg.ir.doc.wrapper import MultiFieldDocumentCollection


class ComponentListener:

    @abstractmethod
    def on_before(self, component, **config):
        print("hook before component running")

    @abstractmethod
    def on_after(self, component, **config):
        print("hook after component running")


class Component:
    def __init__(self, graph_data=None, doc_collection=None):
        if graph_data is not None:
            self.graph_data = graph_data
        else:
            self.graph_data = GraphData()
        if doc_collection is not None:
            self.doc_collection = doc_collection
        else:
            self.doc_collection = MultiFieldDocumentCollection()
        self.__before_run_listeners = []
        self.__after_run_listeners = []

    def set_graph_data(self, graph_data):
        self.graph_data = graph_data

    def set_doc_collection(self, doc_collection):
        self.doc_collection = doc_collection

    def type(self):
        return str(self.__class__.__name__)

    @abstractmethod
    def run(self, **config):
        print("running component %r" % (self.type()))

    def add_before_listener(self, listener: ComponentListener):
        self.__before_run_listeners.append(listener)

    def add_after_listener(self, listener: ComponentListener):
        self.__after_run_listeners.append(listener)

    def before_run(self, **config):
        print("before running component %r" % (self.type()))
        for listener in self.__before_run_listeners:
            listener.on_before(self, **config)

    def after_run(self, **config):
        print("after running component %r" % (self.type()))
        for listener in self.__after_run_listeners:
            listener.on_after(self, **config)

    @abstractmethod
    def provided_entities(self):
        """
        get the entity labels provided after this component
        :return: a list of entity labels
        """
        return set()

    @abstractmethod
    def dependent_entities(self):
        """
        get the dependent entities for the running of this component
        :return: a set of str. each str is the name of the component
        """
        return set()

    @abstractmethod
    def provided_relations(self):
        """
        get the relations provided after this component
        :return: a list of relation names
        """
        return set()

    @abstractmethod
    def dependent_relations(self):
        """
        get the dependent relations for the running of this component
        :return: a set of str. each str is the name of the component
        """
        return set()

    @abstractmethod
    def provided_document_fields(self):
        """
        get the document fields provided after this component, e.g., html_description, clean_short_description
        :return: a list of documents fields
        """
        return set()

    @abstractmethod
    def dependent_document_fields(self):
        """
        get the document fields provided after this component, e.g., html_description, clean_short_description
        :return: a list of documents fields
        """
        return set()



import unittest

from sekg.text.pretreatment.complete_subject import CompleteSubject
from sekg.text.spacy_pipeline.pipeline import PipeLineFactory
from sekg.text.pretreatment.coreference_resolution import ReferenceResolution


class MyTestCase(unittest.TestCase):
    def test_reference_resolution(self):
        r = ReferenceResolution()
        s = """A mutable sequence of characters. This class provides an API compatible with StringBuffer, but with no guarantee of synchronization. This class is designed for use as a drop-in replacement for StringBuffer in places where the string buffer was being used by a single thread (as is generally the case). Where possible, it is recommended that this class be used in preference to StringBuffer as it will be faster under most implementations.
        The principal operations on a StringBuilder are the append and insert methods, which are overloaded so as to accept data of any type. Each effectively converts a given datum to a string and then appends or inserts the characters of that string to the string builder. The append method always adds these characters at the end of the builder; the insert method adds the characters at a specified point.

        For example, if z refers to a string builder object whose current contents are "start", then the method call z.append("le") would cause the string builder to contain "startle", whereas z.insert(4, "le") would alter the string builder to contain "starlet".

        In general, if sb refers to an instance of a StringBuilder, then sb.append(x) has the same effect as sb.insert(sb.length(), x).

        Every string builder has a capacity. As long as the length of the character sequence contained in the string builder does not exceed the capacity, it is not necessary to allocate a new internal buffer. If the internal buffer overflows, it is automatically made larger.

        Instances of StringBuilder are not safe for use by multiple threads. If such synchronization is required then it is recommended that StringBuffer be used."""
        context = {"qn": "java.lang.StringBuilder"}
        remove_result_text, change_list = r.remove_reference_with_context(context, s)
        print(remove_result_text)

        # self.assertEqual(True, False)

    def test_spacy_token(self):
        nlp = PipeLineFactory.full_pipeline()
        query = """Hash table based implementation of the Map interface. This implementation provides all of the optional map operations, and permits null values and the null key. (The HashMap class is roughly equivalent to Hashtable, except that it is unsynchronized and permits nulls.) This class makes no guarantees as to the order of the map; in particular, it does not guarantee that the order will remain constant over time."""
        doc = nlp(query)
        result = []
        for sent in doc.sents:
            result.append(sent.text)
            print(sent.text)
        print(result)

    def test_complete_subject(self):
        complete_subject_tool = CompleteSubject()
        s = """A mutable sequence of characters."""
        context = {"qn": "java.lang.StringBuilder"}
        candidate_verb_set = {"get"}

        s2 = complete_subject_tool.complete_subject_by_name_for_doc(s, candidate_subject_name=context["qn"],
                                                                    candidate_verb_set=candidate_verb_set)
        print(s2)
        s3 = """java.lang.StringBuilder is a mutable sequence of characters."""
        self.assertEqual(s2, s3)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

from sekg.text.extractor.domain_entity.entity_extractor import EntityExtractor


class TestEntityExtractor(TestCase):
    def test_extract_from_sentence(self):
        sent = """Specifically, workspaces are designed for cyclical workloads - such as training neural networks - as
            they allow for off-heap memory reuse (instead of continually allocating and deallocating memory on each iteration
            of the loop). """

        extractor = EntityExtractor()
        terms = extractor.extract_from_sentence(sent)
        for term in terms:
            print(term)

    def test_extract_code_element(self):
        sent = "public void addComponent(String label, Component comp)\n" + \
               "{\n" + \
               "JLabel l = newLabel(label, comp);\n" + \
               "l.setBorder(new EmptyBorder(0,0,0,12));\n" + \
               "addComponent(l,comp,GridBagConstraints.BOTH);\n" + \
               "} //}}\n"
        extractor = EntityExtractor()
        terms = extractor.extract_code_element(sent)
        for term in terms:
            print(term)



import traceback

from py2neo import Subgraph

from sekg.graph.accessor import GraphAccessor


class MetadataGraphAccessor(GraphAccessor):
    """
    query the graph metadata
    """

    # todo: complete this accessor and pass test
    def get_max_id_for_node(self):
        """
            get the max id of Node in the whole graph
            :return: the max id
            """

        query = 'MATCH (n) return max(ID(n))'
        result = self.graph.evaluate(query)
        return result

    def get_max_id_for_relation(self):
        """
        get the max id of relation in the whole graph
        :return: the max id
        """
        query = 'MATCH ()-[n]-() return max(ID(n))'
        result = self.graph.evaluate(query)
        return result

    def get_node_num(self, label=None):
        # todo:
        """
        get the number of the nodes in the graph
        :return: the number of the nodes in the graph
        """
        if label == None:
            query = 'Match (n) RETURN count(n)'
        else:
            query = 'Match (n:`%s`) RETURN count(n)' % (label)
        return self.graph.evaluate(query)

    def get_relation_num(self):
        # todo:
        """
         get the number of the relations in the graph
        :return: the number of the relations in the graph
        """
        query = 'MATCH ()-->() RETURN count(*)'
        return self.graph.evaluate(query)

    def count_relation_type(self):
        """
        get the number of the relation type in the graph
        :return: the number of the relation type in the graph
        """
        query = "MATCH (n)-[r]->() RETURN type(r), count(*)"
        # query = 'CALL apoc.meta.stats() yield relTypeCount'
        result = self.graph.run(query)
        id_num = 0
        for record in result:
            id_num = id_num + 1
        return id_num

    def get_label_set(self):
        pass
        # todo:

    def get_relation_set(self):
        pass
        # todo:

    def expand_node(self, node_id, limit=40):
        """
        get the directly_adjacent_nodes of one node
        :return: return value is a subgraph
        """
        low_quality_query = "Match (n:entity)-[r]-(m:entity) where ID(n)={start_id} return distinct r,n,m limit {limit}"
        low_quality_query = low_quality_query.format(start_id=node_id, limit=limit)
        try:
            nodes = []
            relationships = []
            record_list_for_all_relation = self.graph.run(low_quality_query)

            for record in record_list_for_all_relation:
                r = record["r"]
                relationships.append(r)
                nodes.append(record["n"])
                nodes.append(record["m"])

            if nodes:
                return Subgraph(nodes, relationships)
            else:
                return None
        except Exception:
            traceback.print_exc()
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import itertools
import re
from enum import Enum, unique

# todo: need refactor, some of the relation name are only for the code, eg. mention in
from sekg.text.extractor.domain_entity.nlp_util import SpacyNLPFactory


@unique
class RelationType(Enum):
    IS_A = "is a"
    REPRESENT = "represent"
    SUBCLASS_OF = "subclass of"
    PART_OF = "part of"
    MENTION_IN_COMENT = "mention in comment"
    MENTION_IN_INSIDE_COMENT = "mention in inside comment"
    MENTION_IN_STRING_LITERAL = "mention in string literal"
    MENTION_IN_SHORT_DESCRIPTION = "mention in short description"
    MENTION = "mention"

    HAS_OPERATION = "has operation"
    INSTANCE_OF = "instance of"
    OPERATE = "operate"
    CAN_BE_OPERATED = "can be operated"

    NAME_MENTION = "name mention"


class RelationDetector:
    def __init__(self):
        self.cache = {}
        self.nlp = SpacyNLPFactory.create_spacy_nlp_for_domain_extractor()
        self.pos_cache = {}

    def expand(self, terms):
        terms = set(terms)
        expanded = set()
        addition = set()
        for term in terms:
            tmp = re.sub(r'( and | or )', r' $ ', term)
            parts = tmp.split(" $ ")
            if len(parts) > 1:
                i = 0
                size = len(parts)
                while i < size - 1:
                    p = parts[i].split()
                    n = parts[i + 1].split()
                    if len(p) < len(n):
                        tmp = " ".join(p + n[len(p):])
                        addition.add(tmp)
                        if tmp not in self.cache:
                            self.cache[tmp] = set()
                        self.cache[tmp].add(term)
                        addition.add(parts[i + 1])
                        if parts[i + 1] not in self.cache:
                            self.cache[parts[i + 1]] = set()
                        self.cache[parts[i + 1]].add(term)
                        i += 2
                    else:
                        addition.add(parts[i])
                        if parts[i] not in self.cache:
                            self.cache[parts[i]] = set()
                        self.cache[parts[i]].add(term)
                        i += 1
                        if i == size - 1:
                            addition.add(parts[i])
                            if parts[i] not in self.cache:
                                self.cache[parts[i]] = set()
                            self.cache[parts[i]].add(term)
                expanded.add(term)

            tmp = re.sub(r'( of | \'s )', r' $ ', term)
            parts = tmp.split(" $ ")
            if len(parts) == 2:
                addition.add(parts[0])
                if parts[0] not in self.cache:
                    self.cache[parts[0]] = set()
                self.cache[parts[0]].add(term)
                addition.add(parts[1])
                if parts[1] not in self.cache:
                    self.cache[parts[1]] = set()
                self.cache[parts[1]].add(term)
                expanded.add(term)

        terms.update(addition)
        return expanded

    def detect_relation_by_starfix(self, terms):
        """[summary]

        detect relation by *fix (prefix, infix, suffix)

        Arguments:
            terms {[type]} -- [description]

        Returns:
            [type] -- [description]
        """
        expanded = self.expand(terms)
        relations = set()
        print("terms size:", len(terms))
        candidate = terms - expanded
        new_candidate = set()
        record = dict()
        for c in candidate:
            new_candidate.add(c.lower())
        for i, term in enumerate(new_candidate):
            if i % 10000 == 0:
                print(i)
            for t2 in new_candidate:
                if t2.find(term) >= 0:
                    if term not in record:
                        record[term] = {t2}
                    else:
                        record[term].add(t2)
        print("record")
        for i, term in enumerate(new_candidate):
            if i % 10000 == 0:
                print(i)
            term1 = term
            for term2 in record[term]:
                short_term, long_term = (term1.lower(), term2.lower()) if len(term1) <= len(term2) else (
                    term2.lower(), term1.lower())
                if long_term.find(short_term) < 0:
                    continue
                if long_term.endswith(" " + short_term):
                    if long_term in self.pos_cache:
                        long_term_pos = self.pos_cache[long_term]
                    else:
                        long_term_pos = tuple([token.pos_ for token in self.nlp(long_term)])
                        self.pos_cache[long_term] = long_term_pos
                    if long_term_pos[-1] in {"NOUN", "PROPN"}:
                        short, long = (term1, term2) if len(term1) <= len(term2) else (
                            term2, term1)
                        relations.add((long, RelationType.IS_A.value, short))
                elif long_term.startswith(short_term + " "):
                    if long_term in self.pos_cache:
                        long_term_pos = self.pos_cache[long_term]
                    else:
                        long_term_pos = tuple([token.pos_ for token in self.nlp(long_term)])
                        self.pos_cache[long_term] = long_term_pos
                    if long_term_pos[long_term.split().index(short_term.split()[-1])] in {"NOUN", "PROPN"}:
                        short, long = (term1, term2) if len(term1) <= len(term2) else (
                            term2, term1)
                        relations.add((long, RelationType.PART_OF.value, short))

        for k, v in self.cache.items():
            for term in v:
                relations.add((term, RelationType.PART_OF.value, k))

        relations = set(filter(lambda x: x[0] != x[2], relations))
        relations = self.filter_rules(relations)
        return relations

    def filter_rules(self, relations):
        rel = list(relations)
        for relation in relations:
            r1 = relation[0]
            r2 = relation[2]
            rela = relation[1]
            taga = relation[2] + ' and'
            tagb = 'and ' + relation[2]
            if tagb in relation[0] or taga in relation[0]:
                if relation in rel:
                    rel.append((relation[2], 'consist of', relation[0]))
                    rel.append((relation[0].replace('and', ''), 'part of', relation[2]))
                    rel.remove(relation)
            if 'kind' == relation[2]:
                if relation in rel:
                    rel.remove(relation)
            if 'pair of ' in relation[0]:
                if relation in rel:
                    rel.remove(relation)
            # when ABC -> BC, BC -> C and ABC -> C remove ABC -> C
            for sub_relation in relations:
                if sub_relation[0] == r2 and sub_relation[1] == rela:
                    abcd = (r1, rela, sub_relation[2])
                    if abcd in rel:
                        rel.remove(abcd)
        return set(rel)


if __name__ == "__main__":
    print(RelationType.IS_A.value)



class CodeConstant:
    LABEL_CODE_ELEMENT = "code_element"
    QUALIFIED_NAME = "qualified_name"


class DomainConstant:
    LABEL_DOMAIN_TERM = "domain term"
    PRIMARY_PROPERTY_NAME = "term_name"


class HTMLTYPEConstance:
    HTML_TYPE_UNKNOWN = 0
    HTML_TYPE_API_DECLARATION = 1
    HTML_TYPE_API_SHORT_DESCRIPTION = 2
    HTML_TYPE_API_DETAIL_DESCRIPTION = 3
    HTML_TYPE_METHOD_RETURN_VALUE_DESCRIPTION = 4


class OperationConstance:
    LABEL_OPERATION = "operation"
    PRIMARY_PROPERTY_NAME = "operation_name"


class WikiDataConstance:
    LABEL_WIKIDATA = "wikidata"
    PRIMARY_PROPERTY_NAME = "wd_item_id"
    NAME = "wikidata_name"
    REALTION_SAME_AS = "same as"


class PropertyConstant:
    COMMENT = "comment"
    INSIDE_COMMENT = "inside_comment"
    STRING_LITERAL_EXPR = "string_literal_expr"
    DESCRIPTION = "short_description"
    ALIAS = "alias"
    LEMMA = "lemma"



from sekg.pipeline.base import PipelineListener


class GraphSavePipelineListener(PipelineListener):
    def __init__(self, graph_path):
        self.graph_path = graph_path

    def on_before_run_component(self, component_name, kg_build_pipeline, **config):
        pass

    def on_after_run_component(self, component_name, kg_build_pipeline, **config):
        print("hook after pipeline run component %r" % component_name)
        kg_build_pipeline.save_graph(self.graph_path)


class DocumentSavePipelineListener(PipelineListener):
    def __init__(self, doc_path):
        self.doc_path = doc_path

    def on_before_run_component(self, component_name, kg_build_pipeline, **config):
        pass

    def on_after_run_component(self, component_name, kg_build_pipeline, **config):
        print("hook after pipeline run component %r" % component_name)
        kg_build_pipeline.save_doc(self.doc_path)



import cProfile
import time
import traceback


def conn_try_again(function):
    retries = 5
    # retry time
    count = {"num": retries}

    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as err:
            if count['num'] < 2:
                count['num'] += 1
                return wrapped(*args, **kwargs)
            else:
                raise Exception(err)

    return wrapped


def catch_exception(function):
    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as err:
            traceback.print_exc()

    return wrapped


def exeTime(function):
    def wrapped(*args, **kwargs):
        t0 = time.time()
        print("@%s, {%s} start" % (time.strftime("%X", time.localtime()), function.__name__))
        back = function(*args, **kwargs)
        print("@%s, {%s} end" % (time.strftime("%X", time.localtime()), function.__name__))
        print("@%.3fs taken for {%s}" % (time.time() - t0, function.__name__))
        return back

    return wrapped


def statsProfile(function):
    """
    statics the detail of method run
    :param function:
    :return:
    """

    def wrapped(*args, **kwargs):
        profile = cProfile.Profile()
        profile.enable()
        back = function(*args, **kwargs)
        profile.disable()
        profile.print_stats(sort="tottime")

        return back

    return wrapped



#!/usr/bin/env python
# -*- coding: utf-8 -*-

class SOTagItem:

    def __init__(self, tag_name=None, id=None, count=None, excerpt_post_id=None, wiki_post_id=None,
                 short_description=None, long_description=None, synonyms=None, info_html=None, tag_info_dic={}):
        self.tag_name = tag_name
        self.id = id
        self.count = count
        self.excerpt_post_id = excerpt_post_id
        self.wiki_post_id = wiki_post_id
        self.short_description = short_description
        self.long_description = long_description
        self.synonyms = synonyms
        self.info_html = info_html
        if tag_info_dic:
            self.__init_from_tag_info_dic(tag_info_dic)

    def is_valid(self):
        if (self.short_description or self.long_description) and self.synonyms:
            return True
        return False

    def __init_from_tag_info_dic(self, tag_info_dic):
        if "TagName" in tag_info_dic.keys():
            self.tag_name = tag_info_dic["TagName"]
        if "Id" in tag_info_dic.keys():
            self.id = tag_info_dic["Id"]
        if "Count" in tag_info_dic.keys():
            self.count = tag_info_dic["Count"]
        if "ExcerptPostId" in tag_info_dic.keys():
            self.excerpt_post_id = tag_info_dic["ExcerptPostId"]
        if "WikiPostId" in tag_info_dic.keys():
            self.wiki_post_id = tag_info_dic["WikiPostId"]
        if "ShortDescription" in tag_info_dic.keys():
            self.short_description = tag_info_dic["ShortDescription"]
        if "LongDescription" in tag_info_dic.keys():
            self.long_description = tag_info_dic["LongDescription"]
        if "InfoHtml" in tag_info_dic:
            self.info_html = tag_info_dic["InfoHtml"]

    def get_tag_name(self):
        return self.tag_name

    def get_short_description(self):
        return self.short_description

    def get_long_description(self):
        return self.long_description

    def get_tag_synonyms(self):
        return self.synonyms

    def get_info_html(self):
        return self.info_html

    def get_tag_id(self):
        return self.id

    def get_tag_count(self):
        return self.count

    def get_excerpt_post_id(self):
        return self.excerpt_post_id

    def get_wiki_post_id(self):
        return self.wiki_post_id

    def update_tag_name(self, tag_name):
        self.tag_name = tag_name

    def update_short_description(self, short_description):
        self.short_description = short_description

    def update_long_description(self, long_description):
        self.long_description = long_description

    def update_tag_synonyms(self, synonyms):
        self.synonyms = synonyms

    def update_info_html(self, info_html):
        self.info_html = info_html

    def update_tag_id(self, id):
        self.id = id

    def update_tag_count(self, count):
        self.count = count

    def update_excerpt_post_id(self, excerpt_post_id):
        self.excerpt_post_id = excerpt_post_id

    def update_wiki_post_id(self, wiki_post_id):
        self.wiki_post_id = wiki_post_id



from spacy.tokens.doc import Doc
from spacy.tokens.token import Token
from sekg.text.extractor.domain_entity.identifier_util import IdentifierInfoExtractor

class SentenceHandler:
    __HyphenHandler = 'hyphen_handler'

    # def __init__(self):
    #     self.__hyphen_handler(Doc)
    #     # Doc.set_extension(SentenceHandler.__HyphenHandler, method=SentenceHandler.__hyphen_handler(Doc), default=None, force=True)
    #
    # def __call__(self, doc: Doc):
    #     return self.__hyphen_handler(doc)
    #     # return doc

    @staticmethod
    def hyphen_handler(doc: Doc):
        """
        合并连字符，如fail-fast在tokenize之后会分成fail - fast三个，合并之后变为fail-fast
        :param doc:
        :return:
        """
        hyphen_index = []
        hyphen_index_start = -1
        method_left_brace_cache_index = -1
        left_angle_bracket = -1
        angle_brackets_list = []
        method_brace_pair_list = []
        length = len(doc)
        for i, token in enumerate(doc):
            if '-' == token.text:
                if hyphen_index_start == -1:
                    hyphen_index_start = i - 1
            else:
                if hyphen_index_start != -1:
                    if i != length-1:
                        if doc[i+1].text != '-':
                            hyphen_index.append((hyphen_index_start, i))
                            hyphen_index_start = -1
                        else:
                            continue
                    else:
                        hyphen_index.append((hyphen_index_start, i))
                        hyphen_index_start = -1

        # count = 0
        with doc.retokenize() as retokenizer:
            for start, end in hyphen_index:
                # attrs = {"LEMMA": doc[num - 1].text + '-' + doc[num + 1].text}
                attrs = {"LEMMA": "".join([token.text for token in doc[(start):(end+1)]]), "TAG": doc[end].tag_, "POS": doc[end].pos_}
                retokenizer.merge(doc[start:end+1], attrs=attrs)
                # count += 1
        for i, token in enumerate(doc):
            # todo" complete this for sentence containing method name
            # 添加了对method name的处理
            if "<" == token.text:
                if i > 0 and doc[i-1].text != ">":
                    left_angle_bracket = i-1
                else:
                    left_angle_bracket = i
            elif "<" in token.text:
                left_angle_bracket = i

            if ">" == token.text and left_angle_bracket != -1:
                angle_brackets_list.append((left_angle_bracket, i))
                left_angle_bracket = -1
        # print(method_brace_pair_list)
        with doc.retokenize() as retokenizer:
            for star_index, end_index in angle_brackets_list:
                attrs = {"LEMMA": " ".join([token.text for token in doc[(star_index):(end_index+1)]]),
                         "POS": "NOUN",
                         "TAG": "NN",
                         }
                retokenizer.merge(doc[star_index:end_index+1], attrs=attrs)

        for i, token in enumerate(doc):
            # todo" complete this for sentence containing method name
            # 添加了对method name的处理
            if "(" == token.text:
                if i > 0 and doc[i-1].text != ")" and doc[i-1].text != ".":
                    method_left_brace_cache_index = i-1
                else:
                    method_left_brace_cache_index = i
            elif "(" in token.text:
                method_left_brace_cache_index = i

            if ")" == token.text and method_left_brace_cache_index != -1:
                method_brace_pair_list.append((method_left_brace_cache_index, i))
                method_left_brace_cache_index = -1
        # print(method_brace_pair_list)
        with doc.retokenize() as retokenizer:
            for star_index, end_index in method_brace_pair_list:
                span = doc[star_index:(end_index+1)]
                if not SentenceHandler.is_complete_sentence(span):
                    attrs = {"LEMMA": " ".join([token.text for token in doc[(star_index):(end_index+1)]]),
                             "POS": "NOUN",
                             "TAG": "NN",
                             }
                    retokenizer.merge(doc[star_index:end_index+1], attrs=attrs)

        return doc

    @staticmethod
    def is_complete_sentence(doc):
        root_flag = False
        nusj_flag = False
        for token in doc:
            if token.tag_.startswith('NN'):
                nusj_flag = True
            if token.tag_.startswith('VB'):
                root_flag = True
        if root_flag and nusj_flag:
            return True
        return False

    @staticmethod
    def tag_fixer_after_parser(doc):
        for token in doc:
            if token.dep_ == 'conj' and token.pos_ == 'NOUN' and token.head.pos_ == 'VERB':
                attrs = {"LEMMA": token.lemma_, "TAG": token.head.tag_, "POS": "VERB"}
                with doc.retokenize() as retokenizer:
                    retokenizer.merge(doc[token.i:token.i+1], attrs=attrs)
        return doc

    @staticmethod
    def merge_verb_prep(doc: Doc):
        for token in doc:
            # print(token.text)
            if token.dep_ == 'ROOT' and token.pos_ == 'VERB' and token.tag_ != 'VBN':
                # print(token.text)
                if token.i < len(doc) - 1:
                    if doc[token.i + 1].dep_ == 'prep' and doc[token.i+1].head == token:
                        with doc.retokenize() as retokenizer:
                            attrs = {"LEMMA": " ".join([token.lemma_, doc[token.i+1].lemma_]),
                                     "POS": "VERB",
                                     "TAG": token.tag_
                                     }
                            retokenizer.merge(doc[token.i:token.i + 2], attrs=attrs)
                        if token.i < len(doc) - 1:
                            # print(doc[token.i+1])
                            if doc[token.i+1].dep_ == 'pobj' and doc[token.i+1].pos_ == 'NOUN':
                                with doc.retokenize() as retokenizer:
                                    attrs = {"DEP": "dobj"
                                             }
                                    retokenizer.merge(doc[token.i+1:token.i + 2], attrs=attrs)
                    break
        return doc

    @staticmethod
    def method_name_tag(doc: Doc):
        """
        为method_name自定义词性标注
        :param doc:
        :return:
        """
        idntifier = IdentifierInfoExtractor()
        # name = name.split("(")[0].split(".")[-1]
        # uncamel = self.uncamelize(name).lower()
        # doc = self.nlp(uncamel)
        return idntifier.pos_tag_for_method_name_from_method_doc(doc)

    @staticmethod
    def class_name_tag(doc: Doc):
        """
        为class_name自定义词性标注
        :param doc:
        :return:
        """
        for i in range(0, len(doc)):
            # 因为class_name可能出现RC2StringBuilder这种类型，2,4不一定就是表示to和for
            # if spacy_doc[i].text == "2" and i != 0 and len(spacy_doc) - 1:
            #     spacy_doc[i].lemma_ = "to"
            #     spacy_doc[i].pos_ = "ADP"
            #     spacy_doc[i].tag_ = "IN"
            #
            # if spacy_doc[i].text == "4" and i != 0 and len(spacy_doc) - 1:
            #     spacy_doc[i].lemma_ = "for"
            #     spacy_doc[i].pos_ = "ADP"
            #     spacy_doc[i].tag_ = "IN"

            if doc[i].text.lower() == "to":
                doc[i].lemma_ = "to"
                doc[i].pos_ = "ADP"
                doc[i].tag_ = "IN"

            if doc[i].text.lower() in ["id", "ID", "Id"]:
                doc[i].lemma_ = "id"
                doc[i].pos_ = "NOUN"
                doc[i].tag_ = "NN"

        if doc[-1].pos_ == 'VERB':
            doc[-1].pos_ = 'NOUN'
            doc[-1].tag_ = 'NN'

        return doc



class CodeEntityCategory:
    """
    the basic code type constant, define all things in code we care.
    can be used in model definition and other thing.
    Here we think there are three different category CodeEntity (Container,Type,Value).
    and sometimes one codeEntity could be Container and Type at the same time.

    When possible, you should use the most accurate categories, but in some cases, it is difficult to do so,
    You can use the most suitable type.

    For example, you don't know if an API Entity is an interface, then you can temporarily use TYPE_CLASS as his category,
    and change later.

    Explanation:

    1. Container, the CodeEntity could contain other CodeEntity
        - CATEGORY_PACKAGE, stand for the package, it could contain other class, interface.
        - CATEGORY_CLASS, stand for the class, it could contain method, field.
        - CATEGORY_INTERFACE, stand for the interface, it could contain method.
        - CATEGORY_EXCEPTION_CLASS, stand for the special class- Exception, the Exception class will thrown by method.
            actually TYPE_EXCEPTION_CLASS is subclass of TYPE_EXCEPTION_CLASS.
        - CATEGORY_ERROR_CLASS,  stand for the special class- Error, the Error class will thrown by method.
            actually TYPE_ERROR_CLASS is subclass of TYPE_EXCEPTION_CLASS.
        - CATEGORY_ENUM_CLASS, stand for the special class- ENUM,
        - CATEGORY_ANNOTATION_CLASS, stand for the special class- Annotation class, implement the annotation,

    2. Value, the CodeEntity stand for some value(instance) that has type and special meaning. type is necessary.

        For example, "name:Sting path; type:java.lang.String; description:the path to the file" is a parameter meaning the file path.
        - TYPE_RETURN_VALUE, meaning the return value instance of some methods. Only has the (type,description). and the description sometimes maybe empty.
        - CATEGORY_PARAMETER, meaning the parameter of the method. with type and
        - CATEGORY_FIELD, some field in class or interface, maybe constant or property of class.
            For example, java.awt.Color.RED, (type,name,description,value).
        - CATEGORY_EXCEPTION_CONDITION, description what and when the exception thrown.
        (exception type,description). For example, (type:NullPointerException, thrown when the parameter file path is Null)
        PS: it is not the exception class, it is the exception instance thrown by the method.
        - CATEGORY_LOCAL_VARIABLE, it stands for the local variable in method implement code.

    3. Type:
        - CATEGORY_PRIMARY_TYPE, the primary type for java, etc. int, long, short.
        - CATEGORY_CLASS, stand for the class, it could contain method, field.
        - CATEGORY_INTERFACE, stand for the interface, it could contain method.
        - CATEGORY_EXCEPTION_CLASS, stand for the special class- Exception, the Exception class will thrown by method.
            actually CATEGORY_EXCEPTION_CLASS is subclass of CATEGORY_EXCEPTION_CLASS.
        - CATEGORY_ERROR_CLASS,  stand for the special class- Error, the Error class will thrown by method.
            actually CATEGORY_ERROR_CLASS is subclass of CATEGORY_EXCEPTION_CLASS.
        - CATEGORY_ENUM_CLASS, stand for the special class- ENUM,

    3. Other,
        - CATEGORY_METHOD, the all method, actually the CATEGORY_CONSTRUCT_METHOD is also count for method.
        - CATEGORY_CONSTRUCT_METHOD, special method but also belong to method.
        - CATEGORY_ANNOTATION, some annotation in java. Etc."@Nullable"
        - CATEGORY_XML_ATTRIBUTE, some xml attribute, normally for android. But we think it is not so useful. Maybe we should delete it.
        - CATEGORY_ENUM_CONSTANTS, the constant in CATEGORY_ENUM_CLASS,

        - CATEGORY_BASE_OVERRIDE_METHOD, the parent node for all overriding method without parameters.
        etc. java.util.List.add, java.util.List.append
        - CATEGORY_METHOD_CALL_INSTANCE: the instance for a method call with parameter.
        For example, java.util.List.append(1),java.util.List.append(a)
        - CATEGORY_VALUE, the base type for CATEGORY_PARAMETER,CATEGORY_FIELD,TYPE_RETURN_VALUE
        - CATEGORY_CLASS_FIELD, some special field value, eg.java.awt.AlphaComposite.SrcAtop, java.awt.Color.BLACK.
        Because the extraction, we may not know the value's type, but we has its qualified name and which class that it belong to.

    """
    CATEGORY_UNKNOWN = 0
    CATEGORY_PACKAGE = 1
    CATEGORY_CLASS = 2
    CATEGORY_INTERFACE = 3
    CATEGORY_EXCEPTION_CLASS = 4
    CATEGORY_ERROR_CLASS = 5
    CATEGORY_FIELD = 6
    CATEGORY_CONSTRUCT_METHOD = 7
    CATEGORY_ENUM_CLASS = 8
    CATEGORY_ANNOTATION_CLASS = 9
    CATEGORY_XML_ATTRIBUTE = 10
    CATEGORY_METHOD = 11
    CATEGORY_ENUM_CONSTANTS = 12
    CATEGORY_PRIMARY_TYPE = 13
    CATEGORY_PARAMETER = 14
    CATEGORY_RETURN_VALUE = 15
    CATEGORY_EXCEPTION_CONDITION = 16
    CATEGORY_BASE_OVERRIDE_METHOD = 17
    CATEGORY_VALUE = 18
    CATEGORY_FIELD_OF_CLASS = 19
    CATEGORY_LOCAL_VARIABLE = 20
    CATEGORY_PROPERTY = 21
    CATEGORY_ATTACHED_PROPERTY = 22
    CATEGORY_OPERATOR = 23
    CATEGORY_EVENT = 24
    CATEGORY_STRUCT = 25
    CATEGORY_ATTACHED_EVENT = 26
    CATEGORY_DELEGATE = 27
    CATEGORY_CONSTRUCTORS = 28

    category_code_to_str_map = {
        CATEGORY_UNKNOWN: "unknown",
        CATEGORY_PACKAGE: "package",
        CATEGORY_CLASS: "class",
        CATEGORY_INTERFACE: "interface",
        CATEGORY_EXCEPTION_CLASS: "exception class",
        CATEGORY_ERROR_CLASS: "error class",
        CATEGORY_RETURN_VALUE: "return value",
        CATEGORY_CONSTRUCT_METHOD: "construct method",
        CATEGORY_ENUM_CLASS: "enum class",
        CATEGORY_ANNOTATION_CLASS: "annotation class",
        CATEGORY_XML_ATTRIBUTE: "xml attribute",
        CATEGORY_METHOD: "method",
        CATEGORY_ENUM_CONSTANTS: "enum constants",
        CATEGORY_PRIMARY_TYPE: "primary type",
        CATEGORY_PARAMETER: "parameter",
        CATEGORY_FIELD: "field",
        CATEGORY_EXCEPTION_CONDITION: "exception condition",
        CATEGORY_BASE_OVERRIDE_METHOD: "base override method",
        CATEGORY_VALUE: "value",
        CATEGORY_FIELD_OF_CLASS: "field of class",
        CATEGORY_LOCAL_VARIABLE: "local variable",
        CATEGORY_PROPERTY: "property",
        CATEGORY_ATTACHED_PROPERTY: "attached property",
        CATEGORY_OPERATOR: "operator",
        CATEGORY_EVENT: "event",
        CATEGORY_STRUCT: "struct",
        CATEGORY_ATTACHED_EVENT: "attached event",
        CATEGORY_DELEGATE: "delegate",
        CATEGORY_CONSTRUCTORS: "constructors"
    }

    category_code_to_str_list_map = {
        CATEGORY_UNKNOWN: ["unknown"],
        CATEGORY_PACKAGE: ["package"],
        CATEGORY_CLASS: ["class", "type"],
        CATEGORY_INTERFACE: ["interface", "type"],
        CATEGORY_EXCEPTION_CLASS: ["exception class", "class", "type"],
        CATEGORY_ERROR_CLASS: ["error class", "class", "type"],
        CATEGORY_RETURN_VALUE: ["return value", "value"],
        CATEGORY_CONSTRUCT_METHOD: ["construct method", "method"],
        CATEGORY_ENUM_CLASS: ["enum class", "class", "type"],
        CATEGORY_ANNOTATION_CLASS: ["annotation class", "class"],
        CATEGORY_XML_ATTRIBUTE: ["xml attribute"],
        CATEGORY_METHOD: ["method"],
        CATEGORY_ENUM_CONSTANTS: ["enum constants"],
        CATEGORY_PRIMARY_TYPE: ["primary type", "type"],
        CATEGORY_PARAMETER: ["parameter", "value"],
        CATEGORY_FIELD: ["field", "value"],
        CATEGORY_EXCEPTION_CONDITION: ["exception condition", "value"],
        CATEGORY_BASE_OVERRIDE_METHOD: ["base override method"],
        CATEGORY_VALUE: ["value"],
        CATEGORY_FIELD_OF_CLASS: ["field of class"],
        CATEGORY_LOCAL_VARIABLE: ["local variable", "value"],
        CATEGORY_PROPERTY: ["property"],
        CATEGORY_ATTACHED_PROPERTY: ["attached property"],
        CATEGORY_OPERATOR: ["operator"],
        CATEGORY_EVENT: ["event"],
        CATEGORY_STRUCT: ["struct"],
        CATEGORY_ATTACHED_EVENT: ["attached event"],
        CATEGORY_DELEGATE: ["delegate"],
        CATEGORY_CONSTRUCTORS: ["constructors"]
    }

    JAVA_PRIMARY_TYPE_BYTE = "byte"
    JAVA_PRIMARY_TYPE_CHAR = "char"
    JAVA_PRIMARY_TYPE_SHORT = "short"
    JAVA_PRIMARY_TYPE_INT = "int"
    JAVA_PRIMARY_TYPE_LONG = "long"
    JAVA_PRIMARY_TYPE_FLOAT = "float"
    JAVA_PRIMARY_TYPE_DOUBLE = "double"
    JAVA_PRIMARY_TYPE_BOOLEAN = "boolean"
    JAVA_PRIMARY_TYPE_VOID = "void"

    NET_PRIMARY_TYPE_BYTE = "byte"
    NET_PRIMARY_TYPE_CHAR = "char"
    NET_PRIMARY_TYPE_SHORT = "short"
    NET_PRIMARY_TYPE_USHORT = "ushort"
    NET_PRIMARY_TYPE_DECIMAL = "decimal"
    NET_PRIMARY_TYPE_INT = "int"
    NET_PRIMARY_TYPE_UINT = "uint"
    NET_PRIMARY_TYPE_LONG = "long"
    NET_PRIMARY_TYPE_ULONG = "ulong"
    NET_PRIMARY_TYPE_FLOAT = "float"
    NET_PRIMARY_TYPE_DOUBLE = "double"
    NET_PRIMARY_TYPE_BOOL = "bool"
    NET_PRIMARY_TYPE_SBYTE = "sbyte"
    NET_PRIMARY_TYPE_VOID = "void"

    JAVA_PRIMARY_TYPES = [
        {
            "name": "byte",
            "description": "byte is a keyword which designates the 8 bit signed integer primitive type. The java.lang.Byte class is the nominal wrapper class when you need to store a byte value but an object reference is required."},
        {
            "name": "char",
            "description": "char is a keyword. It defines a character primitive type. char can be created from character literals and numeric representation. Character literals consist of a single quote character (') (ASCII 39, hex 0x27), a single character, and a close quote ('), such as 'w'. Instead of a character, you can also use unicode escape sequences, but there must be exactly one."},
        {
            "name": "short",
            "description": "short is a keyword. It defines a 16 bit signed integer primitive type."},
        {
            "name": "int",
            "description": "int is a keyword which designates the 32 bit signed integer primitive type. The java.lang.Integer class is the nominal wrapper class when you need to store an int value but an object reference is required."
        },
        {
            "name": "long",
            "description": "long is a keyword which designates the 64 bit signed integer primitive type. The java.lang.Long class is the nominal wrapper class when you need to store a long value but an object reference is required."},
        {
            "name": "float",
            "description": "float is a keyword which designates the 32 bit float primitive type. The java.lang.Float class is the nominal wrapper class when you need to store a float value but an object reference is required."},
        {
            "name": "double",
            "description": "double is a keyword which designates the 64 bit float primitive type. The java.lang.Double class is the nominal wrapper class when you need to store a double value but an object reference is required."},
        {
            "name": "boolean",
            "description": "boolean is a keyword which designates the boolean primitive type. There are only two possible boolean values: true and false. The default value for boolean fields is false."},

        {
            "name": "void",
            "description": "void is a Java keyword. Used at method declaration and definition to specify that the method does not return any type, the method returns void. It is not a type and there is no void references/pointers as in C/C++."},
    ]
    NET_PRIMARY_TYPES = [
        {
            "name": "byte",
            "description": "byte is a keyword which designates the 8 bit signed integer primitive type. The java.lang.Byte class is the nominal wrapper class when you need to store a byte value but an object reference is required."},
        {
            "name": "char",
            "description": "char is a keyword. It defines a character primitive type. char can be created from character literals and numeric representation. Character literals consist of a single quote character (') (ASCII 39, hex 0x27), a single character, and a close quote ('), such as 'w'. Instead of a character, you can also use unicode escape sequences, but there must be exactly one."},
        {
            "name": "short",
            "description": "short is a keyword. It defines a 16 bit signed integer primitive type."},
        {
            "name": "int",
            "description": "int is a keyword which designates the 32 bit signed integer primitive type. The java.lang.Integer class is the nominal wrapper class when you need to store an int value but an object reference is required."
        },
        {
            "name": "long",
            "description": "long is a keyword which designates the 64 bit signed integer primitive type. The java.lang.Long class is the nominal wrapper class when you need to store a long value but an object reference is required."},
        {
            "name": "float",
            "description": "float is a keyword which designates the 32 bit float primitive type. The java.lang.Float class is the nominal wrapper class when you need to store a float value but an object reference is required."},
        {
            "name": "double",
            "description": "double is a keyword which designates the 64 bit float primitive type. The java.lang.Double class is the nominal wrapper class when you need to store a double value but an object reference is required."},
        {
            "name": "bool",
            "description": "bool is a keyword which designates the boolean primitive type. There are only two possible boolean values: true and false. The default value for boolean fields is false."},
        {
            "name": "void",
            "description": "void is a Java keyword. Used at method declaration and definition to specify that the method does not return any type, the method returns void. It is not a type and there is no void references/pointers as in C/C++."},
        {
            "name": "ushort",
            "description": "Use the UShort data type to contain binary data too large for Byte. The default value of UShort is 0."},
        {
            "name": "decimal",
            "description": "Holds signed 128-bit (16-byte) values representing 96-bit (12-byte) integer numbers scaled by a variable power of 10. The scaling factor specifies the number of digits to the right of the decimal point; it ranges from 0 through 28."},
        {
            "name": "uint",
            "description": "0 to 4,294,967,295	Unsigned 32-bit integer	System.UInt32"
        },
        {
            "name": "ulong",
            "description": "0 to 18,446,744,073,709,551,615	Unsigned 64-bit integer	System.UInt64"
        },
        {
            "name": "sbyte",
            "description": "Represents a signed 8-bit integer with values ​​ranging from -128 to 127"
        },
    ]
    JAVA_PRIMARY_TYPE_SET = {
        JAVA_PRIMARY_TYPE_BYTE,
        JAVA_PRIMARY_TYPE_CHAR,
        JAVA_PRIMARY_TYPE_SHORT,
        JAVA_PRIMARY_TYPE_INT,
        JAVA_PRIMARY_TYPE_LONG,
        JAVA_PRIMARY_TYPE_FLOAT,
        JAVA_PRIMARY_TYPE_DOUBLE,
        JAVA_PRIMARY_TYPE_BOOLEAN,
        JAVA_PRIMARY_TYPE_VOID,
    }
    NET_PRIMARY_TYPE_SET = {
        NET_PRIMARY_TYPE_BYTE,
        NET_PRIMARY_TYPE_CHAR,
        NET_PRIMARY_TYPE_SHORT,
        NET_PRIMARY_TYPE_USHORT,
        NET_PRIMARY_TYPE_DECIMAL,
        NET_PRIMARY_TYPE_INT,
        NET_PRIMARY_TYPE_LONG,
        NET_PRIMARY_TYPE_ULONG,
        NET_PRIMARY_TYPE_FLOAT,
        NET_PRIMARY_TYPE_DOUBLE,
        NET_PRIMARY_TYPE_BOOL,
        NET_PRIMARY_TYPE_SBYTE,
        NET_PRIMARY_TYPE_UINT,
        NET_PRIMARY_TYPE_VOID,
    }

    @staticmethod
    def java_primary_types():
        return CodeEntityCategory.JAVA_PRIMARY_TYPES

    @staticmethod
    def net_primary_types():
        return CodeEntityCategory.NET_PRIMARY_TYPES

    @staticmethod
    def to_str(category_code):
        if category_code in CodeEntityCategory.category_code_to_str_map:
            return CodeEntityCategory.category_code_to_str_map[category_code]
        return "unknown"

    @staticmethod
    def to_str_list(category_code):
        if category_code in CodeEntityCategory.category_code_to_str_list_map:
            return CodeEntityCategory.category_code_to_str_list_map[category_code]
        return ["unknown"]

    @staticmethod
    def is_basic_type(type_str):
        if type_str in CodeEntityCategory.JAVA_PRIMARY_TYPE_SET:
            return True
        return False

    @staticmethod
    def entity_type_set():
        return CodeEntityCategory.category_code_to_str_map.keys()


class CodeEntityRelationCategory:
    """
    describe different relation between CodeEntity

    - RELATION_CATEGORY_BELONG_TO, the belong to relation between CodeEntity,
         (Method, belongTo, Class/Interface), (Class/Interface, belongTo, Package),
    - RELATION_CATEGORY_EXTENDS, the extends relation between class.
        (Class, extends, Class)
    - RELATION_CATEGORY_IMPLEMENTS, the implement relation between class and interface.
        (Class,implements, Interface)
    - RELATION_CATEGORY_SEE_ALSO, the see also relation between CodeEntity
        eg. (Method, seeAlso, Method)
    - RELATION_CATEGORY_THROW_EXCEPTION_TYPE, between the Exception class and the Method
        (Method, THROW_EXCEPTION_TYPE, Exception class). eg. (java.io.File,THROW_EXCEPTION_TYPE,java.lang.FileNotExistException)
    - RELATION_CATEGORY_RETURN_VALUE_TYPE, between the Class and the Method
        (Method, RETURN_VALUE_TYPE, Exception class). eg. (java.applet.AppletContext.getApplet(java.lang.String),java.applet.Applet)

    - RELATION_CATEGORY_HAS_PARAMETER, between the Method and Parameter value entity,
    - RELATION_CATEGORY_HAS_RETURN_VALUE, between the Method and Return  Value entity,
    - RELATION_CATEGORY_HAS_EXCEPTION_CONDITION, between the Method and EXCEPTION_CONDITION
    - RELATION_CATEGORY_HAS_FIELD, between the Class and field
    - RELATION_CATEGORY_HAS_TYPE, between the Value-CodeEntity and its Type.
    - RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL, some method are call other Method



    - RELATION_CATEGORY_METHOD_OVERRIDING, same method with same parameters in different extended classes
    - RELATION_CATEGORY_METHOD_OVERLOADING, same method with different method parameter in a class
    - RELATION_CATEGORY_METHOD_INSTANCE_CALL, the instance call between method instance and method
    - RELATION_CATEGORY_USE_LOCAL_VARIABLE, the implement code of some method use some local variable. the local variable has name and type.
    """
    RELATION_CATEGORY_BELONG_TO = 1
    RELATION_CATEGORY_EXTENDS = 2
    RELATION_CATEGORY_IMPLEMENTS = 3
    RELATION_CATEGORY_SEE_ALSO = 4

    # relation between method and type
    RELATION_CATEGORY_THROW_EXCEPTION_TYPE = 5
    RELATION_CATEGORY_RETURN_VALUE_TYPE = 6

    # relation between method and the Value-category entity
    RELATION_CATEGORY_HAS_PARAMETER = 7
    RELATION_CATEGORY_HAS_RETURN_VALUE = 8
    RELATION_CATEGORY_HAS_EXCEPTION_CONDITION = 9

    # relation for class and field
    RELATION_CATEGORY_HAS_FIELD = 10
    RELATION_CATEGORY_TYPE_OF = 11

    RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL_METHOD = 13
    RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_USE_CLASS = 14
    RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL_FIELD = 15

    RELATION_CATEGORY_METHOD_OVERRIDING = 16  ## same method with same parameters in different extended classes
    RELATION_CATEGORY_METHOD_OVERLOADING = 17  ## same method with different method parameter in a class

    RELATION_CATEGORY_SUBCLASS_OF = 18
    RELATION_CATEGORY_USE_LOCAL_VARIABLE = 19
    category_code_to_str_map = {
        RELATION_CATEGORY_BELONG_TO: "belong to",
        RELATION_CATEGORY_EXTENDS: "extends",
        RELATION_CATEGORY_IMPLEMENTS: "implements",
        RELATION_CATEGORY_SEE_ALSO: "see also",
        RELATION_CATEGORY_THROW_EXCEPTION_TYPE: "thrown exception type",
        RELATION_CATEGORY_RETURN_VALUE_TYPE: "return value type",
        RELATION_CATEGORY_HAS_PARAMETER: "has parameter",
        RELATION_CATEGORY_HAS_RETURN_VALUE: "has return value",
        RELATION_CATEGORY_HAS_EXCEPTION_CONDITION: "has exception condition",
        RELATION_CATEGORY_HAS_FIELD: "has field",
        RELATION_CATEGORY_TYPE_OF: "type of",
        RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL_METHOD: "call method",
        RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_USE_CLASS: "use class",
        RELATION_CATEGORY_METHOD_IMPLEMENT_CODE_CALL_FIELD: "call field",

        RELATION_CATEGORY_METHOD_OVERRIDING: "overriding",
        RELATION_CATEGORY_METHOD_OVERLOADING: "overloading",

        RELATION_CATEGORY_SUBCLASS_OF: "subclass of",
        RELATION_CATEGORY_USE_LOCAL_VARIABLE: "use local variable"

    }

    @staticmethod
    def to_str(category_code):
        if category_code in CodeEntityRelationCategory.category_code_to_str_map:
            return CodeEntityRelationCategory.category_code_to_str_map[category_code]
        return "unknown"

    @staticmethod
    def relation_set():
        return CodeEntityRelationCategory.category_code_to_str_map.keys()



import operator
import re

from sekg.text.pretreatment.complete_subject import CompleteSubject
from sekg.text.spacy_pipeline.pipeline import PipeLineFactory
from sekg.util.code import CodeElementNameUtil


class ReferenceResolution:
    def __init__(self):
        self.nlp = PipeLineFactory.full_pipeline_for_coreference_resolution()
        self.code_element_name_util = CodeElementNameUtil()

    @staticmethod
    def replace_with_start_end(start, end, word, text):
        """
        根据起点和终点替换字符串
        :param start:
        :param end:
        :param word:
        :param text:
        :return:
        """
        result = text[:start] + word + text[end:]
        return result

    @staticmethod
    def replace_word_in_string(name_replace_list, text):
        """
        根据起始位置、终止位置、替换单词
        :param name_replace_list:
        :param text:
        :return:
        """
        result = text
        name_replace_list.sort(key=operator.itemgetter(0), reverse=True)
        for name_replace in name_replace_list:
            start_pos, end_pos, old_word, new_word = name_replace
            result = ReferenceResolution.replace_with_start_end(start_pos, end_pos, new_word, result)
        return result

    @staticmethod
    def word_num(text):
        """
        有多少个词
        :param text:
        :return: num int
        """
        return len(str(text).split(" "))

    def lemma_word(self, word):
        """
        词性还原
        :param word:
        :return:
        """
        doc = self.nlp(word)
        result = []
        for token in doc:
            if token.lemma_ == "-PRON-":
                result.append(token.norm_)
            else:
                result.append(token.lemma_)
        return " ".join(result)

    def maybe_contain_api(self, name):
        """
        判断name里面是否可能包含API
        :param name:
        :return: True or False
        """
        if name is None or str(name) == "":
            return False

        uncamelize = self.code_element_name_util.uncamelize(name.strip()).replace("( ", "(")
        if len(name.strip().split(" ")) < len(uncamelize.strip().split(" ")) or self.check_upper_case(name):
            return True
        return False

    def check_pronoun(self, word):
        pronoun_set = {
            "it", "that", "this", "these", "those",
            "which", "one", "ones", "they", "other", "another"
        }
        for n in str(word).split(" "):
            if n in pronoun_set:
                return True

        doc = self.nlp(word)
        for token in doc:
            if token.lemma_ == '-PRON-':
                return True

        return False

    @staticmethod
    def replace_ignore_case(text, old, new, ):
        reg = re.compile(re.escape(old), re.IGNORECASE)
        return str(reg.sub(new, text))

    def alias_name_of_qualified_name(self, qn):
        if str(qn).find("(") > 0:
            return qn
        last_part = str(qn).split(".")[-1]
        return self.code_element_name_util.uncamelize(last_part)

    def may_be_api(self, name):
        """
        判断一个名字是不是特别像API
        :param name:
        :return:
        """
        if str(name).strip().find(" ") > 0:
            return False
        if len(str(name).split(".")) >= 3:
            return True

    def remove_reference_with_context(self, context, text, input_doc=None):
        """

        :param input_doc:
        :param context: 一个字典，可以传入方法需要的上下文信息，
        package 信息，class信息，method信息，文档归属对象类型
        :param text: 文本
        :return: text，每句话的改动
        """
        doc_belong_api_name = context["qn"]
        alias = self.alias_name_of_qualified_name(doc_belong_api_name)
        # 按照词数目进行过滤
        skip_size = 4
        may_be_class_set = {"this class ", "the class ", "this interface ", "this abstract class ", "the interface ",
                            "this implementation ", "the methods "}
        my_be_method_set = {"this method ", "the method "}
        if doc_belong_api_name.find("(") < 0:
            for name in may_be_class_set:
                text = ReferenceResolution.replace_ignore_case(text, name, doc_belong_api_name)
        else:
            flag = True
            for my_be_method in my_be_method_set:
                sp_list = text.lower().split(my_be_method)
                if len(sp_list) > 1:
                    ssp_list = sp_list[-1].split(" ")
                    if len(ssp_list) > 0:
                        if ssp_list[0].find("(") < 0:
                            flag = False
                if flag:
                    text = ReferenceResolution.replace_ignore_case(text, my_be_method, doc_belong_api_name + " ")
                    break
        # print(text)
        text = ReferenceResolution.replace_ignore_case(text, "it is recommended that ", "recommended that ")
        text = ReferenceResolution.replace_ignore_case(text, "it will be ", doc_belong_api_name + " will be ")
        if input_doc is None:
            doc = self.nlp(text)
        else:
            doc = input_doc
        row = doc.text
        result = []
        if doc._.coref_clusters is None:
            return row
        all_new_word_list = []
        for cluster in doc._.coref_clusters:
            try:
                new_word = cluster.main.text
                if ReferenceResolution.word_num(new_word) >= skip_size:
                    continue
                for men in cluster.mentions:
                    if cluster.main.text == men.text:
                        continue
                    old_word = men.text
                    start_pos = men.start_char
                    end_pos = men.end_char
                    if new_word == "":
                        continue
                    # 判断old_word是否已经是API了,已经是了就不替换
                    if self.lemma_word(old_word) == self.lemma_word(
                            new_word):
                        continue
                    # 判断old_word是否是代词，不是就continue
                    if not self.check_pronoun(old_word):
                        continue
                    # 判断new_word是否是API
                    if new_word.lower().find(alias.lower()) < 0 and new_word.lower().find(
                            doc_belong_api_name.lower()) < 0:
                        continue

                    all_new_word_list.append(new_word)
                    result.append((start_pos, end_pos, old_word, new_word))
            except Exception as e:
                print(e)
        tmp_after_text = ReferenceResolution.replace_word_in_string(result, row)
        # print(tmp_after_text)
        new_result_list = []
        doc = self.nlp(tmp_after_text)
        for np in doc.noun_chunks:
            for new_word in all_new_word_list:
                pos = str(np.text).find(new_word)
                if pos > 0:
                    new_new_word = ". " + new_word
                    start_pos = np.start_char + pos
                    end_pos = np.end_char
                    new_result_list.append((start_pos, end_pos, new_word, new_new_word))
        after_text = ReferenceResolution.replace_word_in_string(new_result_list, tmp_after_text)

        if doc_belong_api_name.find("(") > 0:
            for my_be_method in my_be_method_set:
                after_text = ReferenceResolution.replace_ignore_case(after_text, my_be_method,
                                                                     " " + doc_belong_api_name + " ")
        else:
            for may_be_class in may_be_class_set:
                after_text = ReferenceResolution.replace_ignore_case(after_text, may_be_class, doc_belong_api_name)

        return after_text, result


    @staticmethod
    def check_upper_case(text):
        """
        判断文本里面是否包含了驼峰式的class
        :param text:
        :return:
        """
        splited = str(text).split(" ")
        for i, s in enumerate(splited):
            if i > 0 and s[0].isupper():
                return True
        return False




#!/usr/bin/env python
# -*- conding: utf-8 -*-
from abc import abstractmethod, ABCMeta
import re
import textdistance


class Rule(metaclass=ABCMeta):
    def __init__(self):
        pass

    @abstractmethod
    def extract(self, sentence):
        pass


class Rule4Bracket(Rule):
    PATTERN = re.compile(r'(\(.*?\))')
    UPPER = set("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

    def __init__(self):
        pass

    def extract(self, text):
        # print(text)
        term_set = set()
        pairs = set()
        for term in self.PATTERN.findall(text):
            cleaned_term = re.sub("'s|’s", "", term)
            cleaned_term = cleaned_term.strip(" '\",.?*/\\‘’()[]")
            index = text.index(term)
            # print(term)
            if index == 0:
                continue
            # abbreviation in bracket
            if len(cleaned_term.split()) == 1:
                if len(cleaned_term) == 1 or len(cleaned_term) > 5:
                    continue
                abbreviation = cleaned_term
                words = text[:index].strip().split()[-(len(cleaned_term) + 1):]

            # full name in bracket
            else:
                abbreviation = text[:index].strip().split()[-1]
                words = cleaned_term.split()
            # print(abbreviation, words)
            upper_count = 0
            for letter in abbreviation:
                if letter in self.UPPER:
                    upper_count += 1
            if upper_count == 0 or len(abbreviation) / upper_count > 2:
                continue
            full_name = self.search(abbreviation, words)
            if full_name is not None:
                term_set.add(abbreviation)
                term_set.add(full_name)
                pairs.add((full_name, abbreviation))

        return term_set, pairs

    def search(self, letters, words):
        temp_letters = letters.lower()
        temp_words = [w.lower() for w in words[:]]

        for i in range(len(temp_words) - 1, -1, -1):
            if temp_words[i][0] == temp_letters[0]:
                # print(temp_words[i:])
                index = self.BFS(temp_letters, temp_words[i:], i)
                if index != -1:
                    name = " ".join(words[i:])
                    if not name.lower().startswith(letters.lower()):
                        return name
        return None

    def BFS(self, letters, words, index):
        if len(letters) == 0:
            return index
        if letters[0] != words[0][0]:
            return -1
        if len(letters) == 1:
            return index + 1
        if len(words) == 1:
            if textdistance.levenshtein(words[0], letters) == len(words[0]) - len(letters):
                return index + 1
            else:
                return -1
        Queue = []
        Queue.append((letters[1:], words[1:]))
        i, j = 1, 1
        for i in range(1, len(letters), 1):
            if j != len(words[0]) and letters[i] in words[0][j:]:
                Queue.append((letters[i + 1:], words[1:]))
            j += 1
        # print(Queue)
        while len(Queue) != 0:
            (l, w) = Queue.pop(0)
            temp = self.BFS(l, w, index + 1)
            # print(temp)
            if temp == -1 and len(Queue) == 0:
                return temp
            elif temp != -1:
                return temp


class Rule4Colon(Rule):
    PATTERN = re.compile(r' ?([a-zA-Z0-9\-]+\+{0,2}) ?:')
    UPPER = set("ABCDEFGHIJKLMNOPQRSTUVWXYZ")

    def __init__(self):
        pass

    def extract(self, text):
        term_set = set()
        pairs = set()
        for term in self.PATTERN.findall(text):
            cleaned_term = re.sub("'s|’s", "", term)
            cleaned_term = cleaned_term.strip(" '\",.?*/\\‘’()[]")
            index = text.index(term)
            if index == len(text) - 1:
                continue
            # abbreviation in bracket
            if len(cleaned_term.split()) == 1:
                if len(cleaned_term) == 1 or len(cleaned_term) > 5:
                    continue
                abbreviation = cleaned_term
                words = text[index + 1:].strip().split()[:len(cleaned_term) + 1]

            upper_count = 0
            for letter in abbreviation:
                if letter in self.UPPER:
                    upper_count += 1
            if upper_count == 0 or len(abbreviation) / upper_count > 2:
                continue
            full_name = self.search(abbreviation, words)

            if full_name is not None:
                term_set.add(abbreviation)
                term_set.add(full_name)
                pairs.add((full_name, abbreviation))
        return term_set, pairs

    def search(self, letters, words):
        temp_letters = letters.lower()
        temp_words = [w.lower() for w in words[:]]

        for i in range(len(temp_words) - 1, -1, -1):
            if temp_words[i][0] == temp_letters[0]:
                # print(temp_words[i:])
                index = self.BFS(temp_letters, temp_words[i:], i)
                if index != -1:
                    name = " ".join(words[i:])
                    if not name.lower().startswith(letters.lower()):
                        return name
        return None

    def BFS(self, letters, words, index):
        if len(letters) == 0:
            return index
        if letters[0] != words[0][0]:
            return -1
        if len(letters) == 1:
            return index + 1
        if len(words) == 1:
            if textdistance.levenshtein(words[0], letters) == len(words[0]) - len(letters):
                return index + 1
            else:
                return -1
        Queue = []
        Queue.append((letters[1:], words[1:]))
        i, j = 1, 1
        for i in range(1, len(letters), 1):
            if j != len(words[0]) and letters[i] in words[0][j:]:
                Queue.append((letters[i + 1:], words[1:]))
            j += 1
        # print(Queue)
        while len(Queue) != 0:
            (l, w) = Queue.pop(0)
            temp = self.BFS(l, w, index + 1)
            # print(temp)
            if temp == -1 and len(Queue) == 0:
                return temp
            elif temp != -1:
                return temp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: isky
@Email: 19110240019@fudan.edu.cn
@Created: 2019/10/28
------------------------------------------
@Modify: 2019/11/18
------------------------------------------
@Description: A file for import the API entity table from the database to the graph
"""
from sekg.constant.code import CodeEntityCategory, CodeEntityRelationCategory
from sekg.constant.constant import CodeConstant
from sekg.graph.builder.code_kg_builder import CodeElementGraphDataBuilder
from sekg.graph.creator import NodeBuilder
from sekg.graph.exporter.graph_data import GraphData

from sekg.pipeline.component.api_importer.model import APIRelation, APIEntity
from sekg.pipeline.component.base import Component


class APIDBImporterComponent(Component):
    LABEL_CODE_ELEMENT = "code_element"
    SUPPORT_LANGUAGE_JAVA = "java"
    SUPPORT_LANGUAGE_C_SHARP = "c#"

    def provided_entities(self):
        entity_labels = set()
        entity_labels.add(APIDBImporterComponent.LABEL_CODE_ELEMENT)
        category_codes = CodeEntityCategory.category_code_to_str_list_map.keys()
        for category_code in category_codes:
            entity_labels.update(set(CodeEntityCategory.to_str_list(category_code)))
        return entity_labels
        # TODO: add interface for the constant

    def dependent_entities(self):
        return set()

    def provided_relations(self):
        relations = set()
        category_codes = CodeEntityRelationCategory.category_code_to_str_map.keys()
        for category_code in category_codes:
            relations.add(CodeEntityRelationCategory.to_str(category_code))
        return relations
        # TODO: add interface for the constant

    def dependent_relations(self):
        return set()

    def provided_document_fields(self):
        return set()

    def dependent_document_fields(self):
        return set()

    # filter_entity_type_set 是一个存放entity type对应数字的集合，把对应的类型过滤掉
    # filter_entity_type_set 是一个存放relation type对应数字的集合，把对应的类型过滤掉

    def __init__(
            self, graph_data=None, session=None, api_entity_table_name=None, api_relation_table_name=None,
            pro_name=None, do_import_primary_type=True, do_infer_extra_relation=True, do_build_aliases=True,
            filter_entity_type_set=None, filter_relation_type_set=None, language=SUPPORT_LANGUAGE_JAVA
    ):
        super().__init__(graph_data)
        if graph_data is None:
            self.graph_data = GraphData()
        self.do_import_primary_type = do_import_primary_type
        self.do_infer_extra_relation = do_infer_extra_relation
        self.do_build_aliases = do_build_aliases
        self.api_entity_table_name = api_entity_table_name
        self.api_relation_table_name = api_relation_table_name
        self.session = session
        self.api_entity_type = None
        self.api_relation_type = None
        self.pro_name = pro_name
        self.code_element_kg_builder = CodeElementGraphDataBuilder(self.graph_data)
        self.language = language
        if filter_entity_type_set is not None:
            self.filter_entity_type_set = filter_entity_type_set
        else:
            self.filter_entity_type_set = set()
        if filter_relation_type_set is not None:
            self.filter_relation_type_set = filter_relation_type_set
        else:
            self.filter_relation_type_set = set()

    def run(self, **config):
        print("running component %r" % (self.type()))
        print("import API entities from table %r" % self.api_entity_table_name)
        print("import API entity relations from table %r" % self.api_relation_table_name)
        self.api_entity_type = APIEntity
        self.api_entity_type.__table__.name = self.api_entity_table_name
        self.api_relation_type = APIRelation
        self.api_relation_type.__table__.name = self.api_relation_table_name
        if self.do_import_primary_type:
            self.import_primary_type()
        api_id_to_node_id_map = self.import_entity_from_api_table()
        self.import_relation_from_table(self.session, api_id_to_node_id_map)
        if self.do_infer_extra_relation:
            self.infer_extra_relation()
        if self.do_build_aliases:
            self.build_aliases()
        self.add_source_label(self.pro_name)

    def before_run(self, **config):
        super().before_run(**config)
        print("before running component %r" % (self.type()))

    def add_source_label(self, source_label):
        """
        给来自不同项目的API打上不同的标签
        :param source_label:
        :return:
        """
        self.code_element_kg_builder.add_source_label(source_label)

    def import_entity_from_api_table(self):
        """
        从数据库表中读取所有的API实体，添加到graphData当中
        """
        print("start api from table")
        api_entity_list = self.session.query(self.api_entity_type).all()
        api_id_to_node_id_map = {}
        for i, entity_info_row in enumerate(api_entity_list):
            api_entity_json = dict(entity_info_row.__dict__)
            api_entity_json.pop('_sa_instance_state', None)
            api_id = api_entity_json["id"]
            qualified_name = api_entity_json[CodeConstant.QUALIFIED_NAME]
            api_type = api_entity_json["api_type"]
            # 过滤掉一些不想要的api类型
            if api_type in self.filter_entity_type_set:
                continue

            normal_entity_types = {
                CodeEntityCategory.CATEGORY_CLASS,
                CodeEntityCategory.CATEGORY_PACKAGE,
                CodeEntityCategory.CATEGORY_METHOD,
                CodeEntityCategory.CATEGORY_INTERFACE,
                CodeEntityCategory.CATEGORY_EXCEPTION_CLASS,
                CodeEntityCategory.CATEGORY_ENUM_CLASS,
                CodeEntityCategory.CATEGORY_ERROR_CLASS,
                CodeEntityCategory.CATEGORY_ANNOTATION_CLASS,
            }
            node_id = GraphData.UNASSIGNED_NODE_ID
            if api_type in normal_entity_types:
                node_id = self.import_normal_entity(api_entity_json)
            if api_type == CodeEntityCategory.CATEGORY_CONSTRUCT_METHOD:
                node_id = self.import_construct_method_entity(api_entity_json)

            if api_type == CodeEntityCategory.CATEGORY_FIELD_OF_CLASS:
                node_id = self.import_qualified_field_entity(api_entity_json)

            if api_type == CodeEntityCategory.CATEGORY_ENUM_CONSTANTS:
                node_id = self.import_qualified_enum_constants_entity(api_entity_json)
            if api_type == CodeEntityCategory.CATEGORY_PRIMARY_TYPE:
                node_id = self.add_primary_type(primary_type_name=qualified_name, **api_entity_json)

            if api_type == CodeEntityCategory.CATEGORY_PARAMETER:
                node_id = self.import_parameter_entity(api_entity_json)
            if api_type == CodeEntityCategory.CATEGORY_RETURN_VALUE:
                node_id = self.import_return_value_entity(api_entity_json)
            if api_type == CodeEntityCategory.CATEGORY_EXCEPTION_CONDITION:
                node_id = self.import_exception_condition_entity(api_entity_json)
            if node_id == GraphData.UNASSIGNED_NODE_ID:
                print("Adding fail %d %s %r " % (api_id, qualified_name, CodeEntityCategory.to_str(api_type)))
                continue
            api_id_to_node_id_map[api_id] = node_id

        self.graph_data.print_graph_info()
        print("end import from api_table ")

        return api_id_to_node_id_map

    def import_relation_from_table(self, session, api_id_to_node_id_map):
        """
        :param session: 数据库连接session
        :param api_id_to_node_id_map: 数据库api_id到nodeId的映射map
        :return:
        """
        print("start import relation")
        self.graph_data.print_graph_info()

        valid_api_types = CodeEntityRelationCategory.relation_set()
        valid_api_types = valid_api_types - self.filter_relation_type_set
        for relation_type in valid_api_types:
            relation_str = CodeEntityRelationCategory.to_str(relation_type)
            print("start import_test relation %s" % (relation_str))
            api_relation_list = session.query(self.api_relation_type).filter(
                self.api_relation_type.relation_type == relation_type).all()
            for relation in api_relation_list:
                if relation.start_api_id not in api_id_to_node_id_map:
                    print("start_id %d can't found its node id" % (relation.start_api_id))
                    continue
                if relation.end_api_id not in api_id_to_node_id_map:
                    print("end_id %d can't found its node id" % (relation.end_api_id))
                    continue
                self.graph_data.add_relation(startId=api_id_to_node_id_map[relation.start_api_id],
                                             endId=api_id_to_node_id_map[relation.end_api_id],
                                             relationType=relation_str)
        print("end import relation")
        self.graph_data.print_graph_info()

    def after_run(self, **config):
        super().before_run(**config)
        print("after running component %r" % (self.type()))

    def import_normal_entity(self, api_entity_json):

        format_qualified_name = self.code_element_kg_builder.format_qualified_name(
            api_entity_json[CodeConstant.QUALIFIED_NAME])

        if not format_qualified_name:
            return
        api_entity_json.pop(CodeConstant.QUALIFIED_NAME)
        node_id = self.code_element_kg_builder.add_normal_code_element_entity(format_qualified_name,
                                                                              api_entity_json["api_type"],
                                                                              **api_entity_json)
        return node_id

    def import_parameter_entity(self, api_entity_json):
        extra_properties = {}

        qualified_name = api_entity_json[CodeConstant.QUALIFIED_NAME]
        short_description = api_entity_json["short_description"]

        value_type = qualified_name.split(" ")[0].strip()
        value_name = qualified_name.split(" ")[1].strip()
        node_id = self.code_element_kg_builder.add_base_value_entity_node(value_type=value_type, value_name=value_name,
                                                                          short_description=short_description,
                                                                          entity_category=CodeEntityCategory.CATEGORY_PARAMETER,
                                                                          **extra_properties)

        if node_id == GraphData.UNASSIGNED_NODE_ID:
            print("fail to add parameter node %r" % (api_entity_json))

        return node_id

    def import_return_value_entity(self, api_entity_json):
        extra_properties = {}

        qualified_name = api_entity_json[CodeConstant.QUALIFIED_NAME]
        short_description = api_entity_json["short_description"]

        value_type = qualified_name.split(" ")[0].strip()
        node_id = self.code_element_kg_builder.add_base_value_entity_node(value_type=value_type, value_name="<R>",
                                                                          short_description=short_description,
                                                                          entity_category=CodeEntityCategory.CATEGORY_RETURN_VALUE,
                                                                          **extra_properties)

        if node_id == GraphData.UNASSIGNED_NODE_ID:
            print("fail to add parameter node %r" % (api_entity_json))

        return node_id

    def import_exception_condition_entity(self, api_entity_json):
        extra_properties = {}

        qualified_name = api_entity_json[CodeConstant.QUALIFIED_NAME]
        short_description = api_entity_json["short_description"]

        value_type = qualified_name.split(" ")[0].strip()
        node_id = self.code_element_kg_builder.add_base_value_entity_node \
            (value_type=value_type, value_name="<E>",
             short_description=short_description,
             entity_category=CodeEntityCategory.CATEGORY_EXCEPTION_CONDITION,
             **extra_properties)

        if node_id == GraphData.UNASSIGNED_NODE_ID:
            print("fail to add parameter node %r" % (api_entity_json))

        return node_id

    def import_construct_method_entity(self, api_entity_json):

        format_qualified_name = self.code_element_kg_builder.format_qualified_name(
            api_entity_json[CodeConstant.QUALIFIED_NAME])

        method_name = self.code_element_kg_builder.parse_construct_to_javaparser_style(format_qualified_name)

        if not method_name:
            return GraphData.UNASSIGNED_NODE_ID

        api_entity_json.pop(CodeConstant.QUALIFIED_NAME)
        node_id = self.code_element_kg_builder.add_normal_code_element_entity(method_name,
                                                                              api_entity_json["api_type"],
                                                                              **api_entity_json)

        return node_id

    def import_qualified_field_entity(self, api_entity_json):

        qualified_name = self.code_element_kg_builder.format_qualified_name(
            api_entity_json[CodeConstant.QUALIFIED_NAME])
        if not qualified_name:
            print(
                "import_qualified_field_entity %r %r" % (api_entity_json[CodeConstant.QUALIFIED_NAME], api_entity_json))
            return GraphData.UNASSIGNED_NODE_ID

        api_entity_json.pop(CodeConstant.QUALIFIED_NAME)
        api_entity_json.pop("api_type")

        node_id = self.code_element_kg_builder.add_normal_code_element_entity(qualified_name,
                                                                              CodeEntityCategory.CATEGORY_FIELD_OF_CLASS,
                                                                              **api_entity_json)

        return node_id

    def import_qualified_enum_constants_entity(self, api_entity_json):

        qualified_name = self.code_element_kg_builder.format_qualified_name(
            api_entity_json[CodeConstant.QUALIFIED_NAME])
        if not qualified_name:
            print(
                "import_qualified_field_entity %r %r" % (api_entity_json[CodeConstant.QUALIFIED_NAME], api_entity_json))
            return GraphData.UNASSIGNED_NODE_ID

        api_entity_json.pop(CodeConstant.QUALIFIED_NAME)
        api_entity_json.pop("api_type")

        node_id = self.code_element_kg_builder.add_normal_code_element_entity(qualified_name,
                                                                              CodeEntityCategory.CATEGORY_ENUM_CONSTANTS,
                                                                              **api_entity_json)

        return node_id

    def add_primary_type(self, primary_type_name, **properties):
        """
        添加一些默认的基础数据类型
        """
        properties[CodeConstant.QUALIFIED_NAME] = primary_type_name

        cate_labels = CodeEntityCategory.to_str_list(CodeEntityCategory.CATEGORY_PRIMARY_TYPE)
        builder = NodeBuilder()
        builder = builder.add_property(**properties).add_entity_label().add_labels(
            APIDBImporterComponent.LABEL_CODE_ELEMENT, *cate_labels)
        node_id = self.graph_data.add_node(
            node_id=GraphData.UNASSIGNED_NODE_ID,
            node_labels=builder.get_labels(),
            node_properties=builder.get_properties(),
            primary_property_name=CodeConstant.QUALIFIED_NAME)
        return node_id

    def import_primary_type(self):
        # todo: this is a not wrong. should be a config of the component, language.
        type_list = []
        if self.language == APIDBImporterComponent.SUPPORT_LANGUAGE_JAVA:
            type_list = CodeEntityCategory.java_primary_types()
        if self.language == APIDBImporterComponent.SUPPORT_LANGUAGE_C_SHARP:
            type_list = CodeEntityCategory.net_primary_types()

        for item in type_list:
            code_element = {
                CodeConstant.QUALIFIED_NAME: item["name"],
                "api_type": CodeEntityCategory.CATEGORY_PRIMARY_TYPE,
                "short_description": item["description"]
            }
            self.add_primary_type(item["name"], **code_element)

        print(self.graph_data)
        self.graph_data.print_label_count()

    def build_aliases(self):
        """
        为API添加别名信息
        """
        self.code_element_kg_builder.build_aliases_for_code_element()
        self.graph_data.refresh_indexer()

    def infer_extra_relation(self):
        self.code_element_kg_builder.build_belong_to_relation()
        self.code_element_kg_builder.build_abstract_overloading_relation()
        self.code_element_kg_builder.build_belong_to_relation()
        self.code_element_kg_builder.build_override_relation()

    def save(self, path):
        self.graph_data.save(path)



from urllib.parse import unquote


class URLUtil:
    @staticmethod
    def parse_url_to_title(url):
        site_str = url.replace('https://en.wikipedia.org/wiki/', '')
        title = site_str.strip(" _").replace("_", " ").replace('"', "&quot;")
        title = unquote(str(title))
        return title



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sekg.graph.exporter.graph_data import GraphData

graph_data = GraphData()

graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear"})

# print(graph_data.find_nodes_by_property_value_starts_with("qualified_name", "ArrayList"))

# print(graph_data.get_node_ids())
# print(graph_data.get_relation_pairs_with_type())

graph_data.add_relation(1, "related to", 2)
graph_data.add_relation(1, "related to", 3)
graph_data.add_relation(1, "related to", 4)
graph_data.add_relation(2, "related to", 3)
graph_data.add_relation(3, "related to", 4)

print(graph_data.get_relations(1, "related to"))
print("get relation by type")
print(graph_data.get_relations(relation_type="related to"))

graph_data.remove_node(node_id=1)

# print(graph_data.get_node_ids())
# print(graph_data.get_relation_pairs_with_type())

# print("#" * 50)
# graph_data.merge_two_nodes_by_id(1, 2)

# print(graph_data.get_node_ids())
# print(graph_data.get_relation_pairs_with_type())

print("full graph----------")
graph_data.print_graph_info()


print("subgraph----------")
left_graph=graph_data.subgraph(set([1,2,3]))
left_graph.print_graph_info()


print("old graph----------")
graph_data.print_graph_info()




from abc import abstractmethod

"""
fuse each concept with concept in Wikidata.
"""


class GenericKGFusion:
    @abstractmethod
    def fuse(self):
        """
        fuse the given term with the concept in wikidata. e.g. link 'JSON' to "(JavaScript Object Notation) https://www.wikidata.org/wiki/Q2063"
        """
        pass



# -*- coding: utf-8 -*-
import json
import traceback

import requests


class WikiDataItem:

    def __init__(self, wd_item_id, init_at_once=True):
        self.wd_item_id = wd_item_id
        self.property_name_list = None
        self.data_dict = None
        self.relation_property_name_list = None
        self.is_init = False
        if init_at_once:
            self.init_wikidata_item(self.wd_item_id)

    def get_en_name(self):
        if self.data_dict:
            return self.data_dict.get("labels_en", "")
        return ""

    def get_en_aliases(self):
        if self.data_dict:
            return self.data_dict.get("aliases_en", [])
        return []

    def get_en_description(self):
        if self.data_dict:
            return self.data_dict.get("descriptions_en", "")
        return ""

    @staticmethod
    def __merge_value_from_dict_list(dict_list, value_key='value'):
        values = []
        for d in dict_list:
            if value_key in d:
                values.append(d[value_key])
        # return ',,'.join(values)
        return values

    def __parse_wd_json_to_dict(self, wd_item_json):
        wd_item_property_dict = {}
        self.__extract_muilti_language_property(wd_item_json, 'aliases', wd_item_property_dict)
        self.extract_muilti_language_property_from_dict(wd_item_json, 'labels', wd_item_property_dict)
        self.extract_muilti_language_property_from_dict(wd_item_json, 'descriptions', wd_item_property_dict)

        self.__extract_sitelinks(wd_item_json, wd_item_property_dict)
        property_name_list, relation_property_name_list = WikiDataItem.__extract_claims(wd_item_json=wd_item_json,
                                                                                        wd_item_property_dict=wd_item_property_dict)
        if 'labels_en' in wd_item_property_dict:
            wd_item_property_dict['name'] = wd_item_property_dict['labels_en']
        return wd_item_property_dict, property_name_list, relation_property_name_list

    @staticmethod
    def __extract_metadata(entity_metadata):
        metadata_dict = {}
        metadata_dict['wd_item_id'] = entity_metadata['id']
        metadata_dict['lastrevid'] = entity_metadata['lastrevid']
        metadata_dict['modified'] = entity_metadata['modified']
        return metadata_dict

    @staticmethod
    def __extract_sitelinks(wd_item_json, wd_item_property_dict):
        if 'sitelinks' in wd_item_json.keys():
            sitelinks = wd_item_json['sitelinks']
            if not sitelinks:
                return
            if "enwiki" not in sitelinks:
                return
            wd_item_property_dict["site:enwiki"] = "https://en.wikipedia.org/wiki/{}".format(
                sitelinks["enwiki"]['title'])
            # for k, v in sitelinks.items():
            #     k_title = "site:" + k
            #     wd_item_property_dict[k_title] = "https://{}.wikipedia.org/wiki/{}".format(v['site'][:-4], v['title'])

    @staticmethod
    def __extract_claims(wd_item_json, wd_item_property_dict):
        claims_dict = wd_item_json['claims']
        if not claims_dict:
            return [], []
        property_name_list = []
        relation_property_name_list = []
        for k, v in claims_dict.items():
            property_name_list.append(k)
            wd_item_property_dict[k] = WikiDataItem.__extract_claims_property_from_value_list(v)
            if v and WikiDataItem.is_property_relation(v[0]):
                relation_property_name_list.append(k)
        return property_name_list, relation_property_name_list

    @staticmethod
    def __extract_claims_property_from_value_list(value_list):
        extract_values = []
        for value in value_list:
            t = WikiDataItem.__extract_claims_property_item_from_value_list(value)
            if t:
                extract_values.append(t)
        # return ',,'.join(extract_values)
        return extract_values

    @staticmethod
    def is_property_relation(value_dict):
        try:
            mainsnak = value_dict['mainsnak']
            if mainsnak['snaktype'] == 'novalue' or mainsnak['snaktype'] == 'somevalue':
                return False
            data_type = mainsnak['datatype']
            if data_type == 'wikibase-item':
                return True
            else:
                return False
        except Exception as e:
            traceback.print_exc()
            return False

    @staticmethod
    def __extract_claims_property_item_from_value_list(value_dict):
        mainsnak = value_dict['mainsnak']
        if mainsnak['snaktype'] == 'novalue' or mainsnak['snaktype'] == 'somevalue':
            return None
        data_type = mainsnak['datatype']
        value = ''

        string_value_type = ['string', 'external-id', 'math', 'url', 'commonsMedia', 'geo-shape']

        if data_type in string_value_type:
            value = mainsnak['datavalue']['value']
        if data_type == 'wikibase-item':
            value = mainsnak['datavalue']['value']['id']
        if data_type == 'wikibase-property':
            value = 'P' + str(mainsnak['datavalue']['value']['numeric-id'])
        if data_type == 'time':
            value = mainsnak['datavalue']['value']['time']
        if data_type == 'monolingualtext':
            value = mainsnak['datavalue']['value']['text']
        if data_type == 'quantity':
            value = mainsnak['datavalue']['value']['amount']
        if data_type == 'globe-coordinate':
            value = str(mainsnak['datavalue']['value']['latitude']) + ',' + str(
                mainsnak['datavalue']['value']['longitude'])

        return value

    def __extract_muilti_language_property(self, wd_item_json, key, wd_item_property_dict):
        language_dict = wd_item_json[key]
        if not language_dict:
            return
        if "en" not in language_dict:
            return
        try:
            wd_item_property_dict[key + "_en"] = self.__merge_value_from_dict_list(language_dict["en"])
        except Exception as error:
            print(error)
        # for k, v in language_dict.items():
        #     try:
        #         full_property_name = key + '_' + k
        #         wd_item_property_dict[full_property_name] = self.__merge_value_from_dict_list(v)
        #     except Exception as error:
        #         print(error)

    @staticmethod
    def extract_muilti_language_property_from_dict(wd_item_json, key, wd_item_property_dict):
        language_dict = wd_item_json[key]
        if not language_dict:
            return
        if "en" not in language_dict:
            return
        try:
            wd_item_property_dict[key + "_en"] = language_dict["en"]['value']
        except Exception as error:
            print(error)
        # for k, v in language_dict.items():
        #     try:
        #         full_property_name = key + '_' + k
        #         wd_item_property_dict[full_property_name] = v['value']
        #     except Exception as error:
        #         print(error)

    def init_wikidata_item_from_json_string(self, data_json_string):
        try:

            dict_json = self.parse_illegal_json_string(data_json_string)

            self.wd_item_id = dict_json["id"]
            self.__parse(dict_json)
            self.is_init = True
            return self
        except Exception as error:
            traceback.print_exc()
            self.is_init = False
            return self

    def init_wikidata_item_from_json(self, json_data):
        self.__parse(json_data["entities"][self.wd_item_id])
        self.is_init = True
        return self

    def init_wikidata_item(self, wd_item_id):
        try:
            r = requests.get(
                "https://www.wikidata.org/wiki/Special:EntityData/{wd_item_id}.json".format(wd_item_id=wd_item_id))
            json_response = r.content.decode(encoding='utf-8')
            dict_json = json.loads(json_response, encoding='utf-8')
            self.wd_item_id = wd_item_id
            self.__parse(dict_json["entities"][wd_item_id])
            self.is_init = True
            return self
        except Exception as error:
            traceback.print_exc()
            self.is_init = False
            return self

    def init_wikidata_item_from_wikipedia_title(self, wikipedia_title):
        try:
            r = requests.get(
                "https://en.wikipedia.org/w/api.php?action=query&prop=pageprops&ppprop=wikibase_item&redirects=1&format=json&titles=" + wikipedia_title)

            json_response = r.content.decode(encoding='utf-8')
            query_json = json.loads(json_response, encoding='utf-8')
            wd_item_id = None
            pages = query_json["query"]["pages"]
            for page_index, page in pages.items():
                if page_index == "-1":
                    break
                wd_item_id = page["pageprops"]["wikibase_item"]
            self.wd_item_id = wd_item_id
            if wd_item_id is None:
                self.is_init = False
                return self
            return self.init_wikidata_item(wd_item_id=wd_item_id)

        except Exception as error:
            traceback.print_exc()
            self.is_init = False
            return self

    def init_wikidata_item_from_wikipedia_url(self, wikipedia_url):
        wikipedia_title = wikipedia_url.replace("https://en.wikipedia.org/wiki/", "")
        return self.init_wikidata_item_from_wikipedia_title(wikipedia_title)

    def __parse(self, source_wd_dict_json):
        property_dict, property_name_list, relation_property_name_list = self.__parse_wd_json_to_dict(
            source_wd_dict_json)
        metadata_dict = self.__extract_metadata(source_wd_dict_json)
        data_dict = dict(property_dict, **metadata_dict)
        self.data_dict = data_dict
        self.property_name_list = property_name_list
        self.relation_property_name_list = relation_property_name_list

    def get_wikidata_item_property_dict(self):
        if self.data_dict == None:
            return {}
        return self.data_dict

    def get_relation_property_name_list(self):
        if self.relation_property_name_list == None:
            return []
        return self.relation_property_name_list

    def get_wikidata_item_property_name_list(self):
        if self.property_name_list == None:
            return []
        return self.property_name_list

    def get_non_relation_property_name_list(self):
        non_relation_property_set = set(self.get_wikidata_item_property_name_list()) - set(
            self.get_relation_property_name_list())
        return list(non_relation_property_set)

    def exist(self):
        return self.is_init

    def get_en_wiki_url(self):
        if self.data_dict:
            return self.data_dict.get("site:enwiki", None)
        return None

    def get_en_wiki_title(self):
        if self.data_dict and "site:enwiki" in self.data_dict:
            return self.data_dict["site:enwiki"].split("/")[-1]
        return None

    @staticmethod
    def parse_illegal_json_string(data_json_string):

        if data_json_string is None or data_json_string == "":
            return None

        try:
            try:
                json_instance = json.loads(data_json_string, encoding='utf-8')
                return json_instance
            except Exception:
                data_json_string = json.dumps(eval(data_json_string))

                json_instance = json.loads(data_json_string, encoding='utf-8')
                return json_instance
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def is_valid_json_string(data_json_string):
        try:
            json_instance = WikiDataItem.parse_illegal_json_string(data_json_string)
            if json_instance == None:
                return False
            if json_instance["claims"] == {}:
                return False
            return True
        except Exception:
            traceback.print_exc()
            return False

    # todo: add a toString method

    def get_neighbor_ids(self):
        neighbor_ids = set()
        for r in self.relation_property_name_list:
            id_set = set([])
            end = self.data_dict.get(r, [])
            if type(end) == list:
                for e in end:
                    id_set.add(e)
            else:
                id_set.add(end)
            neighbor_ids.update(id_set)
        return neighbor_ids



from unittest import TestCase

from sekg.graph.accessor import GraphAccessor
from sekg.graph.exporter.graph_data import GraphData, Neo4jImporter
from sekg.graph.factory import GraphInstanceFactory

import networkx


class TestGraphData(TestCase):
    test_config_path = "neo4j_config.json"
    test_server_id = 4

    def test_get_graph(self):
        graph_data = GraphData()

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear"})

        # print(graph_data.find_nodes_by_property_value_starts_with("qualified_name", "ArrayList"))

        # print(graph_data.get_node_ids())
        # print(graph_data.get_relation_pairs_with_type())

        graph_data.add_relation(1, "related to", 2)
        graph_data.add_relation(1, "related to", 3)
        graph_data.add_relation(1, "related to", 4)
        graph_data.add_relation(2, "related to", 3)
        graph_data.add_relation(3, "related to", 4)

        print(graph_data.get_relations(1, "related to"))
        print("get relation by type")
        print(graph_data.get_relations(relation_type="related to"))

        # print(graph_data.get_node_ids())
        # print(graph_data.get_relation_pairs_with_type())

        # print("#" * 50)
        # graph_data.merge_two_nodes_by_id(1, 2)

        # print(graph_data.get_node_ids())
        # print(graph_data.get_relation_pairs_with_type())

    def test_get_graph_with_property(self):
        graph_data = GraphData()

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear"})

        # print(graph_data.find_nodes_by_property_value_starts_with("qualified_name", "ArrayList"))

        # print(graph_data.get_node_ids())
        # print(graph_data.get_relation_pairs_with_type())

        graph_data.add_relation_with_property(1, "related to", 2, extra_info_key="as")
        graph_data.add_relation_with_property(1, "related to", 3, extra_info_key="ab")
        graph_data.add_relation_with_property(1, "related to", 4, extra_info_key="cs")
        graph_data.add_relation_with_property(2, "related to", 3, extra_info_key="ca")
        graph_data.add_relation_with_property(3, "related to", 4)

        print(graph_data.get_relations(1, "related to"))
        print("get relation by type")
        print(graph_data.get_relations(relation_type="related to"))
        t = graph_data.get_edge_extra_info(1, 2, "related to", extra_key="extra_info_key")
        print(t)
        # print(graph_data.get_node_ids())
        # print(graph_data.get_relation_pairs_with_type())

        # print("#" * 50)
        # graph_data.merge_two_nodes_by_id(1, 2)

        # print(graph_data.get_node_ids())
        # print(graph_data.get_relation_pairs_with_type())

    def test_merge(self):
        graph_data = GraphData()
        graph_data.create_index_on_property("qualified_name", "alias")

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear", "alias": ["clear"]})

        graph_data.merge_node(node_labels=["method", "merge"], node_properties={"qualified_name": "ArrayList.clear",
                                                                                "alias": ["clear", "clear1"]
                                                                                },
                              primary_property_name="qualified_name")

    def test_find_nodes_by_properties(self):
        graph_data = GraphData()
        graph_data.create_index_on_property("qualified_name", "alias")

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear", "alias": ["clear"]})
        graph_data.add_node({"method"},
                            {"qualified_name": "List.clear", "alias": ["clear", "List.clear", "List clear"]})

        match_nodes = graph_data.find_nodes_by_property(property_name="qualified_name", property_value="List.clear")
        print(match_nodes)
        self.assertIsNotNone(match_nodes)
        self.assertEqual(len(match_nodes), 1)
        self.assertEqual(match_nodes[0][GraphData.DEFAULT_KEY_NODE_ID], 5)

        match_nodes = graph_data.find_nodes_by_property(property_name="alias", property_value="clear")

        print(match_nodes)
        self.assertIsNotNone(match_nodes)
        self.assertEqual(len(match_nodes), 2)

    def test_graph_data_2_neo4j(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = GraphAccessor(graph=graph)
        neo4j_importer = Neo4jImporter(accessor)
        graph_data = GraphData()
        graph_data.create_index_on_property("qualified_name", "alias")

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear", "alias": ["clear"]})
        graph_data.add_node({"method"},
                            {"qualified_name": "List.clear", "alias": ["clear", "List.clear", "List clear"]})
        graph_data.add_relation(1, "related to", 2)
        graph_data.add_relation(1, "related to", 3)
        graph_data.add_relation(1, "related to", 4)
        graph_data.add_relation(2, "related to", 3)
        graph_data.add_relation(3, "related to", 4)
        t = networkx.shortest_path(graph_data.graph, 2, 4)
        print(graph_data.get_relations(1, "related to"))
        print("get relation by type")
        print(graph_data.get_relations(relation_type="related to"))
        neo4j_importer.import_all_graph_data(graph_data)
        pass

    def test_remove_node(self):
        graph_data = GraphData()
        graph_data.create_index_on_property("qualified_name", "alias")

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add"})
        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear", "alias": ["clear"]})
        graph_data.add_node({"method"},
                            {"qualified_name": "List.clear", "alias": ["clear", "List.clear", "List clear"]})
        graph_data.add_relation(1, "related to", 2)
        graph_data.add_relation(1, "related to", 3)
        graph_data.add_relation(1, "related to", 4)
        graph_data.add_relation(2, "related to", 3)
        graph_data.add_relation(3, "related to", 4)

        result = graph_data.remove_node(node_id=1)
        self.assertIsNotNone(result)

        self.assertIsNone(graph_data.get_node_info_dict(node_id=1))



# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: attenton
@Email: 18212010081@fudan.edu.cn
@Created: 2019/12/13
------------------------------------------
@Modify: 2019/12/13
------------------------------------------
@Description:
"""
import re

import spacy
from spacy.lang.char_classes import ALPHA, HYPHENS


class CustomizeSpacy:
    """
    apcy中的tokenizer主要原理分为五个部分，例外rules（也可以被称为tokenizer_exceptions）， 前缀prefix，后缀suffix，中缀infix，以及例外的正则匹配token_match
    修改以上任意一个都可以修改spacy中的tokenizer，实现自定义的tokenizer
    """
    @staticmethod
    def customize_tokenizer_split_single_lowcase_letter_and_period(nlp):
        """
        自定义的tokenzier，修改spacy中的exception，
        spacy中的tokenizer exception添加了{a.|b.|c.}等单个字母与句号连接的例外，这个可能导致分句出现错误，比如 1 < n < r. 出现在句尾。
        :param nlp:
        :return:
        """
        # prefix_re = spacy.util.compile_prefix_regex(nlp.Defaults.prefixes)
        # suffix_re = spacy.util.compile_suffix_regex(nlp.Defaults.suffixes)
        # infix_re = spacy.util.compile_infix_regex(nlp.Defaults.infixes)
        #
        # # remove all exceptions where a single letter is followed by a period (e.g. 'h.')
        exceptions = {k: v for k, v in dict(nlp.Defaults.tokenizer_exceptions).items() if
                      not (len(k) == 2 and k[1] == '.')}
        # new_tokenizer = Tokenizer(nlp.vocab, exceptions,
        #                           prefix_search=prefix_re.search,
        #                           suffix_search=suffix_re.search,
        #                           infix_finditer=infix_re.finditer,
        #                           token_match=nlp.tokenizer.token_match)
        #
        # nlp.tokenizer = new_tokenizer
        nlp.tokenizer.rules = exceptions

    @staticmethod
    def customize_tokenizer_merge_hyphen(nlp):
        """
        将连字符相连的合并为一个token， 如原来的fail-fast在tokenize之后会分成fail - fast三个，修改之后会全部合并为fail-fast这一个token
        实现方法： 删除infix中分离连字符的正则表达式
        :param nlp:
        :return:
        """
        _infixes = list(nlp.Defaults.infixes)
        _infixes.remove(r"(?<=[{a}])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS),)
        infix_regex = spacy.util.compile_infix_regex(_infixes)
        nlp.tokenizer.infix_finditer = infix_regex.finditer

    @staticmethod
    def customize_tokenizer_api_name_recognition(nlp):
        """
        将句子中的api名字识别成一个token, 在生成tokenizer的函数后加上token.match
        TODO: 目前只是将中括号的合并了，由于token_match是在suffix、prefix之后执行，所以复杂的正则会被suffix、prefixoverride，所以目前还是用doc.retoeknizer()方法实现api_name的识别
        :param nlp:
        :return:
        """
        # re_token_match = _get_regex_pattern(nlp.Defaults.token_match)
        _suffixes = list(nlp.Defaults.suffixes)
        _suffixes.remove("\]")
        suffixes_regex = spacy.util.compile_suffix_regex(_suffixes)
        nlp.tokenizer.suffix_search = suffixes_regex.search
        # api_name_match = re.compile("(\w+(\[[\w+-]+\])?\.)+\w+\(.*?\)", re.UNICODE).match
        # token_match = nlp.Defaults.token_match
        # print(token_match)
        # token_match.append(api_name_match)
        # nlp.tokenizer.token_match = api_name_match

if __name__ == '__main__':
    nlp = spacy.load('en_core_web_sm')
    CustomizeSpacy.customize_tokenizer_split_single_lowcase_letter_and_period(nlp)
    CustomizeSpacy.customize_tokenizer_merge_hyphen(nlp)
    CustomizeSpacy.customize_tokenizer_api_name_recognition(nlp)
    # sentence = 'java.lang.StringBuffer create StringBuffer'
    sentence = 'Suppose that a byte sequence of length n is written, where < n <= r. Up to the first srcs[offset].remaining() bytes of this sequence are written from buffer srcs[offset], up to the next srcs[offset+1].remaining() bytes are written from buffer srcs[offset+1], and so forth, until the entire byte sequence is written.'
    # sentence = "The elements in the array returned by the method executeBatch may be one of the following: A number greater than or equal to zero -- indicates that the command was processed successfully and is an update count giving the number of rows in the database that were affected by the command's execution"
    # sentence = "The maximum number of buffers-dke-dfs to be accessed 1 < n < r. So API must be non-negative and no larger than srcs.length - offset Returns long The number of bytes written, possibly zero"
    doc = nlp(sentence)
    # debug tokenizer
    tok_exp = nlp.tokenizer.explain(sentence)
    for t in tok_exp:
        print(t[1], "\t", t[0])
    for i in doc.sents:
        print(i)
        print([token.text for token in i])
        print([token.dep_ for token in i])
        print([token.tag_ for token in i])
        print([token.tag_ for token in i])
        print('\n')


from sekg.ir.preprocessor.spacy import SpacyTextPreprocessor

if __name__ == "__main__":
    preprocessor = SpacyTextPreprocessor()
    test_case_list = [
        (
            "This is a ArrayList, it contains all of the classes for creating user interfaces and for painting graphics and images."
            , ['ArrayList',
               'class',
               'user',
               'interface',
               'painting',
               'graphic',
               'image',
               'contain',
               'create']
        ),

        (
            "how to get File's MD5 checksum",
            ['File', 'md5', 'checksum']

        )

    ]

    for old_str, new_str in test_case_list:
        team = preprocessor.extract_words_for_query(old_str)



import spacy

from nltk.corpus import stopwords
import nltk
import re
from nltk.stem import WordNetLemmatizer

from sekg.ir.preprocessor.base import Preprocessor
from sekg.graph.util.name_searcher import KGNameSearcher

simplenlp = spacy.load('en', disable=["ner", "parser"])
nlp = spacy.load('en')


# todo: the preprocess could change to almost same as extract domain Entity_NP, by fixing the tokenize problem

class SpacyTextPreprocessor(Preprocessor):
    def __init__(self, kg_name_searcher=None):
        self.name_searcher = kg_name_searcher
        self.lemmatizer = WordNetLemmatizer()

    def extract_concept_and_operation(self, query):
        np_list = []
        v_list = []
        # remove stop_word
        new_query = query
        # 处理引号问题
        words = re.split(r"\"|'", new_query)
        if len(words) % 2 == 0:
            words.pop(len(words) - 1)
        for index, np in enumerate(words):
            if index % 2 == 1 and len(str(np)) > 0:
                lemma_token = self.lemmatizer.lemmatize(np)
                # 将得到的特殊词汇进行替换 替换为NP
                new_query = new_query.replace(str("\"" + np + "\""), "Entity_NP")
                np_list.append(lemma_token)
        del words
        doc = nlp(new_query)
        for np in doc.noun_chunks:
            lemma_token = np.lemma_
            # 将得到的名词进行替换 替换为NP
            if lemma_token != "Entity_NP":
                new_query = new_query.replace(str(np), "Entity_NP")
                lemma_token = " ".join(
                    [w for w in nltk.word_tokenize(lemma_token) if w not in stopwords.words('english')])
                lemma_token = lemma_token.replace("Entity_NP", "")
                lemma_token = lemma_token.strip()
                if len(lemma_token) > 0:
                    np_list.append(lemma_token)
        # 相邻的np进行合并，合并新的np
        if len(np_list) > 0:
            np_position = [[index, item] for index, item in enumerate(new_query.split()) if item == "Entity_NP"]
            index = np_position[0][0] + 1
            for i in range(1, len(np_position)):
                if index == np_position[i][0]:
                    np_list[i - 1] += (" " + np_list[i])
                    np_list.pop(i)
                    index += 1
                else:
                    index = np_position[i][0] + 1
        doc = nlp(query)
        for token in doc:
            if token.pos_ == "VERB" and token.is_stop is False:
                # 提取动词
                verb_lemma = token.lemma_
                v_list.append(verb_lemma)
            elif token.pos_ != "NOUN" and token.is_stop is False:
                # 提取其他词汇之中含有同名概念或者操作节点的短语， 若是概念短语，则定义为concept，否则是operation
                token_lemma = token.lemma_
                if len(self.name_searcher.search_by_full_name(token_lemma)) > 0 \
                        and token_lemma not in stopwords.words('english'):
                    tag = False
                    for item in np_list:
                        if token_lemma in item:
                            tag = True
                            break
                    if not tag:
                        np_list.append(token_lemma)
        new_np_list = []
        for item in np_list:
            pattern = re.findall("(\s+[^a-zA-Z1-90]+\s+)", item)
            for p in pattern:
                item = item.replace(p, " ")
            pattern = re.findall("([^a-zA-Z1-90]+\s+)", item)
            for p in pattern:
                item = item.replace(p, " ")
            pattern = re.findall("(\s+[^a-zA-Z1-90]+)", item)
            for p in pattern:
                item = item.replace(p, " ")
            new_np_list.append(item)
        np_list = new_np_list
        del new_np_list
        # 删除特殊字符
        np_list = [re.sub("([^a-zA-Z1-90-]+)", " ", item, count=0, flags=0).strip() for item in np_list]
        np_list = [item for item in np_list if len(re.findall("([a-z|A-Z|1-9|0]+)", item)) > 0]
        return np_list, v_list

    def extract_words_for_query(self, query):
        np_v_list = []

        doc = nlp(query)
        for np in doc.noun_chunks:
            np_words = []
            for np_token in np:
                lemma_token = np_token.lemma_
                if np_token.is_stop:
                    continue

                if np_token.is_digit == True:
                    continue

                if np_token.is_punct == True:
                    continue

                np_words.append(lemma_token)

            clean_np_str = " ".join(np_words)
            if clean_np_str:
                np_v_list.append(clean_np_str)

        for token in doc:
            if token.is_stop == True:
                continue
            if token.is_digit == True:
                continue

            if token.is_punct == True:
                continue

            if token.pos_ == "VERB" and token.is_stop == False:
                verb_lemma = token.lemma_
                np_v_list.append(verb_lemma)

        # todo: extract CamelName split as keyword from this

        return set(" ".join(np_v_list).split())

    def clean(self, text):
        """
        return a list of token from the text, only remove stopword and lemma the word
        :param text: the text need to preprocess
        :return: list of str
        """

        result = []
        doc = simplenlp(text)

        for token in doc:
            if token.is_stop is True:
                continue
            if token.is_digit is True:
                continue

            if token.is_punct is True:
                continue

            result.append(token.lemma_.lower())

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import re

import textdistance
from kgtools.annotation import TimeLog, Cache
from nltk.corpus import wordnet
from spacy.lang.en import LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES
from spacy.lemmatizer import Lemmatizer

from sekg.term.synset import Synset
from sekg.util.annotation import exeTime


class Fusion:
    def __init__(self):
        self.DL = textdistance.DamerauLevenshtein()
        self.lemmatizer = Lemmatizer(LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES)

    @Cache
    def get_synonyms_from_wordnet(self, word):
        """
        全部同义词集合
        :param word:
        :return:
        """
        #todo: has problem. need to fixed
        synonyms = []
        for syn in wordnet.synsets(word):
            for l in syn.lemmas():
                synonyms.append(l.name())
        print("word=",word,"syms=",synonyms)
        return set(synonyms)

    def get_simple_type_for_term(self, term):

        seperate_words = [" of ", " Of "]

        # A of B type
        for s_w in seperate_words:
            if s_w in term:
                words = term.split(s_w)
                if len(words) != 2:
                    return term
                child = words[0]
                parent = words[1]
                return (parent + " " + child).replace("  ", " ")

        seperate_words = ["'s ", "' "]

        # A's B => A B
        for s_w in seperate_words:
            if s_w in term:
                words = term.split(s_w)
                if len(words) != 2:
                    return term
                parent = words[0]
                child = words[1]

                return (parent + " " + child).replace("  ", " ")

        seperate_words = ["[][]", "[] []", "[ ] [ ]"]

        for s_w in seperate_words:
            if s_w in term:
                words = term.split(s_w)
                if len(words) != 2:
                    return term
                parent = words[0]
                child = words[1]

                return ("array of " + parent + " array " + child).replace("  ", " ").strip()

        seperate_words = ["[]", "[ ]"]

        for s_w in seperate_words:
            if s_w in term:
                words = term.split(s_w)
                if len(words) != 2:
                    return term
                parent = words[0]
                child = words[1]

                return ("array of " + parent + " " + child).replace("  ", " ").strip()

        return term

    @Cache
    def clean_name_for_synonym(self, term):
        term = self.lemmatizer(term, 'NOUN')[0].lower()
        term = re.sub(r'[-/_]', "", term)

        term = re.sub(r"\s+", " ", term)
        term = self.get_simple_type_for_term(term)
        return term

    def check_synonym(self, long_term, short_term, max_dis=2, threshold=1 / 5):

        long_name = self.clean_name_for_synonym(long_term)
        short_name = self.clean_name_for_synonym(short_term)

        if long_name == short_name:
            return True
        if long_name.replace(" ", "") == short_name.replace(" ", ""):
            return True
        if long_term.lower().rstrip("s") == short_term.lower().rstrip("s"):
            return True

        if long_term.isupper() and short_term.isupper():
            return False

        long_words, short_words = long_name.split(), short_name.split()

        if len(long_words) != len(short_words):
            return False
        for word1, word2 in zip(long_words, short_words):
            if word1[0] != word2[0]:
                return False
            if re.findall(r'[0-9]+', word1) != re.findall(r'[0-9]+', word2):
                return False
            if self.compute_DL_distance(word1, word2) > max_dis or self.compute_DL_distance_normalize(word1,
                                                                                                      word2) >= threshold:
                return False
        return True

    @Cache
    def compute_DL_distance_normalize(self, word1, word2):
        return self.DL.normalized_distance(word1, word2)

    @Cache
    def compute_DL_distance(self, word1, word2):
        return self.DL.distance(word1, word2)

    def check_abbr(self, long_term, short_term, stopwords=None):
        def __check_prefix(long_name, short_name):
            if len(short_name) >= 2 and len(short_name) / len(long_name) < 2 / 3 and long_name.startswith(short_name):
                return True
            return False

        def __check_phrase_word(phrase, word, stopwords):
            long_words = phrase.split()
            if len(word) < len(long_words):
                return False
            if len(long_words) == 1:
                return __check_word_word(phrase, word, stopwords)
            else:
                all_pairs = itertools.combinations([i for i in range(1, len(word), 1)], len(long_words) - 1)
                queue = []
                for pair in all_pairs:

                    temp = []
                    pair = (0,) + pair + (len(word),)
                    # print(pair)
                    long_word = long_words[0]
                    segment = word[pair[0]:pair[1]]
                    if long_word[0] == segment[0] or (
                            len(long_word) > 1 and long_word[0] in set("aeiou") and segment[0] == long_word[1]):
                        temp.append((long_word, segment))
                    else:
                        continue
                    for i in range(1, len(pair) - 1):
                        long_word = long_words[i]
                        segment = word[pair[i]:pair[i + 1]]
                        if long_word[0] == segment[0]:
                            temp.append((long_word, segment))
                        else:
                            break
                    else:
                        queue.append(temp)
                # print(queue)
                for pairs in queue:
                    for long_word, segment in pairs:
                        # print(segment)
                        if not __check_word_word(long_word, segment, stopwords):
                            return False
                        if len(segment) > 1 and segment != "re":
                            return False
                    else:
                        return True
                return False

        def __check_word_word(long_word, word, stopwords):
            if (len(word) == 1 and (word[0] == long_word[0])) or __check_prefix(long_word, word):
                return True
            return False

        def __in(short_name, long_name):
            beg = 0
            length = len(long_name)
            for ch in short_name:
                index = long_name.find(ch, beg)
                if index == -1:
                    return False
                beg = index + 1
                if beg >= length:
                    return False
            return True

        long_name = self.lemmatizer(long_term, 'NOUN')[0].lower()
        short_name = self.lemmatizer(short_term, 'NOUN')[0].lower()
        long_name, short_name = re.sub(r'[-/]', "", long_name), re.sub(r'[-/]', "", short_name)
        long_name, short_name = re.sub(r"\s+", " ", long_name), re.sub(r"\s+", " ", short_name)

        if not __in(short_name, long_name):
            return False

        # print(long_name, short_name)

        if len(short_name.split()) == 1:
            return __check_phrase_word(long_name, short_name, stopwords)
        else:
            short_words = short_name.split()
            long_words = long_name.split()
            if len(long_words) < len(short_words):
                return False
            while len(short_words) > 0 and short_words[0] == long_words[0]:
                short_words.pop(0)
                long_words.pop(0)
            while len(short_words) > 0 and short_words[-1] == long_words[-1]:
                short_words.pop(-1)
                long_words.pop(-1)
            if len(short_words) == 1:
                # print(long_words, short_words)
                return __check_phrase_word(" ".join(long_words), short_words[0], stopwords)
            else:
                return False

    def handle_pairs(self, pair_list, stopwords):
        inner_synonyms = set()
        inner_abbreviations = set()
        for t1, t2 in pair_list:
            # print("%s, %s" % (t1.text, t2.text))
            if len(t1) == 1 or len(t2) == 1:
                continue
            short_term, long_term = (t1, t2) if len(t1) <= len(t2) else (t2, t1)
            if self.check_synonym(long_term, short_term):
                inner_synonyms.add((long_term, short_term))
            elif self.check_abbr(long_term, short_term, stopwords):
                inner_abbreviations.add((short_term, long_term))
        return inner_synonyms, inner_abbreviations

    def detect_relations(self, terms, stopwords):
        pairs = [(t1, t2) for t1, t2 in itertools.combinations(terms, 2)]
        return self.handle_pairs(pairs, stopwords)

    def merge_synonyms(self, synonyms_relations):
        queue = [set(r) for r in synonyms_relations]

        # print(queue)
        # print("Merge synonyms")

        def __merge(queue):
            synonyms = []
            done = True
            while len(queue) > 0:
                now = queue.pop(0)
                # print("now:", now)
                merged = []
                for r in queue:
                    if len(r & now) > 0:
                        now |= r
                        merged.append(r)
                        done = False
                for r in merged:
                    queue.remove(r)
                synonyms.append(now)
            if done:
                return synonyms
            else:
                return __merge(synonyms)

        synonyms = set([Synset(synonym) for synonym in __merge(queue)])
        return synonyms

    def merge_abbrs(self, abbr_relations, synonyms, threshold=0.5):
        term2synset = {}
        for synset in synonyms:
            for term in synset.terms:
                term2synset[term] = synset

        abbr2fullname = {}
        for start, end in abbr_relations:
            if start not in term2synset:
                term2synset[start] = Synset(set([start]))
            if end not in term2synset:
                term2synset[end] = Synset(set([end]))
            start_synset = term2synset[start]
            end_synset = term2synset[end]
            if start_synset not in abbr2fullname:
                abbr2fullname[start_synset] = set()
            abbr2fullname[start_synset].add(end_synset)

        abbrs = set()
        for abbr, fullnames in abbr2fullname.items():
            for fn in fullnames:
                fn += abbr
        synsets = set(term2synset.values())
        return synsets

    def handle_pair_for_synonym(self, t1, t2):
        inner_synonyms = set()

        # print("%s, %s" % (t1.text, t2.text))
        if len(t1) == 1 or len(t2) == 1:
            return inner_synonyms
        short_term, long_term = (t1, t2) if len(t1) <= len(t2) else (t2, t1)
        if self.check_synonym(long_term, short_term):
            inner_synonyms.add((long_term, short_term))
        #     wordnet 可以融合
        # if self.get_synonyms_from_wordnet(short_term) & self.get_synonyms_from_wordnet(long_term):
        #     inner_synonyms.add((long_term, short_term))
        return inner_synonyms

    def handle_pairs_for_synonym(self, pair_list, stopwords):
        inner_synonyms = set()
        for t1, t2 in pair_list:
            inner_synonyms = inner_synonyms | self.handle_pair_for_synonym(t1, t2)
        return inner_synonyms

    def handle_term_to_list_for_synonym(self, start_term, end_term_list):
        inner_synonyms = set()
        for end_term in end_term_list:
            inner_synonyms = inner_synonyms | self.handle_pair_for_synonym(start_term, end_term)
        return inner_synonyms

    def detect_synonym_relations(self, terms, stopwords):
        result = set()
        terms = list(terms)
        for start_term_index, start_t in enumerate(terms):
            end_term_list = terms[(start_term_index + 1):]
            result = result | self.handle_term_to_list_for_synonym(start_t, end_term_list)

        for t in terms:
            lemma = self.lemmatizer(t, 'NOUN')[0].lower()
            if lemma != t:
                result.add((t, lemma))
        return result

    @exeTime
    def fuse_by_synonym(self, terms, term_count=None, stopwords=None):
        if type(terms) != set:
            terms = set(terms)
        synonym_relations = self.detect_synonym_relations(terms, stopwords)
        # print("syn: ", [(t1.text, t2.text) for t1, t2 in synonym_relations])
        # print("abbr:", [(t1.text, t2.text) for t1, t2 in abbr_relations])
        synsets = self.merge_synonyms(synonym_relations)

        visited = set()
        for synset in synsets:
            visited.update(synset.terms)
        for term in terms - visited:
            synsets.add(Synset({term}))

        if term_count is not None:
            for synset in synsets:
                synset.init_count(term_count)
        return synsets

    @TimeLog
    def fuse(self, terms, term_count=None, stopwords=None):

        synonym_relations, abbr_relations = self.detect_relations(terms, stopwords)
        # print("syn: ", [(t1.text, t2.text) for t1, t2 in synonym_relations])
        # print("abbr:", [(t1.text, t2.text) for t1, t2 in abbr_relations])
        synonyms = self.merge_synonyms(synonym_relations)
        synsets = self.merge_abbrs(abbr_relations, synonyms)

        visited = set()
        for synset in synsets:
            visited.update(synset.terms)
        for term in terms - visited:
            synsets.add(Synset({term}))

        if term_count is not None:
            for synset in synsets:
                synset.init_count(term_count)
        return synsets



from unittest import TestCase

from sekg.ir.preprocessor.ngram import NGramPreprocessor
from sekg.graph.util.name_searcher import KGNameSearcher


class TesNGramPreprocessor(TestCase):

    def test_get_n_gram_list(self):
        name_searcher = KGNameSearcher.load(
            "E:\PycharmProjects\KGFeatureLocation\output\graph\jedite-4.3\\NameSearcher\jedite-4.3.v4.2.namesearcher"
        )
        preprocessor = NGramPreprocessor(4, kg_name_searcher=name_searcher)
        query_str = [
            "folding bug, text is in a black hole",
            "Size of file open/save dialogs incorrect on dual displays",
            "Incorrect Python code indentation",
            "Search expression",
            "Autosave turned off for Untitled Documents",
            "Keyboard navigation through menubar inserts char into buffer",
            "Add filter to recent files menu",
            "File Open Dialog: can't handle full pathnames",
            "Provide a way of reloading changed buffers without prompting",
            "Rectangular Selection and Dragging",
            "Filter Bug in File System Browser",
            "Auto-indent breaks if previous line is a single line indent",
            "set \"Search for:\" to \"||\" check \"Regular expressions\" click \"Replace All\" version : jEdit 4.3pre8",
            "Wrong conversion to UTF-16 (with BOM)",
            "caret/display position w/ multiple views open",
        ]
        for query in query_str:
            print(preprocessor.get_n_gram_list(query))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: isky
@Email: 19110240019@fudan.edu.cn
@Created: 2019/10/16
------------------------------------------
@Modify: 2019/10/16
------------------------------------------
@Description:
"""
import unittest
from unittest import TestCase

from sekg.graph.exporter.graph_data import GraphData
from sekg.graph.util.name_searcher import KGNameSearcher


class TestKGNameSearcher(TestCase):
    def get_graph(self):
        graph_data = GraphData()

        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add()", "name": "ArrayList.add",
                                         "alias": ["ArrayList.add1", "add()", "add"]})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.add(int)", "name": "ArrayList.add",
                                         "alias": ["ArrayList.add2", "add()", "add"]})

        graph_data.add_node({"override method"}, {"qualified_name": "ArrayList.pop"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.remove"})
        graph_data.add_node({"method"}, {"qualified_name": "ArrayList.clear"})

        return graph_data

    def test_train_by_names_properties(self):
        graph_data = self.get_graph()
        searcher = KGNameSearcher()
        searcher.train_by_names_properties(graph_data, "name", "qualified_name", "alias")

        self.assertEqual(set([1]), searcher.search_by_full_name("ArrayList.add()"))
        self.assertEqual(set([2]), searcher.search_by_full_name("ArrayList.add(int)"))
        self.assertEqual(set([1,2]), searcher.search_by_full_name("ArrayList.add"))
        self.assertEqual(set([2]), searcher.search_by_full_name("ArrayList.add2"))


if __name__ == '__main__':
    unittest.main()



from nltk.corpus import wordnet as wn
from sekg.text.extractor.domain_entity.word_util import WordUtil


# from word_forms.word_forms import get_word_forms


class VocabularyConversion:
    def __init__(self):
        self.WN_NOUN = 'n'
        self.WN_VERB = 'v'
        self.WN_ADJECTIVE = 'a'
        self.WN_ADJECTIVE_SATELLITE = 's'
        self.WN_ADVERB = 'r'

    # def word_forms_find(self, word):
    #     forms = get_word_forms(word)
    #     print(forms)
    #     return forms
    @staticmethod
    def couldBeVerb_probability(word):
        """
        check if a word could be a verb
        :param word:
        :return:
        """
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if "v" in tag_count:
            return True
        return False

    @staticmethod
    def couldBeADJECTIVE_High_probability(word):
        return VocabularyConversion.couldBeHigh_probability(word, "a")

    @staticmethod
    def couldBeADJECTIVE_SATELLITE_High_probability(word):
        return VocabularyConversion.couldBeHigh_probability(word, "s")

    @staticmethod
    def couldBeNOUN_High_probability(word):
        return VocabularyConversion.couldBeHigh_probability(word, "n")

    @staticmethod
    def couldBeADVERB_High_probability(word):
        return VocabularyConversion.couldBeHigh_probability(word, "r")

    @staticmethod
    def couldbe_adj(word):
        verb_type = "a"
        if word is None:
            return False
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if verb_type == "a" and "s" in tag_count:
            if "a" in tag_count:
                tag_count["a"] += tag_count["s"]
            else:
                tag_count["a"] = tag_count["s"]
        if verb_type in tag_count and tag_count[verb_type] > 0:
            return True
        return False

    @staticmethod
    def couldBeHigh_probability(word, verb_type, confidence=0.5):
        """
        :param verb_type:
        :param word:
        :return:
        """
        if word is None:
            return False
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if verb_type == "a" and "s" in tag_count:
            if "a" in tag_count:
                tag_count["a"] += tag_count["s"]
            else:
                tag_count["a"] = tag_count["s"]
        if verb_type in tag_count and tag_count[verb_type] >= (len(pairs) * confidence):
            return True
        return False

    @staticmethod
    def non_more_than_adj(word):
        if word.endswith("ed") or (word.endswith("able") and (not word.endswith("table"))):
            return False
        if word is None:
            return False
        sp_list = word.split()
        if len(sp_list) > 1:
            word = sp_list[-1]
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if "s" in tag_count:
            if "a" in tag_count:
                tag_count["a"] += tag_count["s"]
            else:
                tag_count["a"] = tag_count["s"]
        if "n" not in tag_count:
            return False
        if "a" not in tag_count:
            return True
        if tag_count["n"] > tag_count["a"]:
            return True
        return False

    def noun_2_verb(self, word):
        """
        名词转动词
        :param word:
        :return:
        """
        t = self.convert(word, from_pos=self.WN_VERB, to_pos=self.WN_NOUN)
        if len(t) != 0 and t[0][1] > 0.5 and t[0][0] != "string":
            return word, True
        if self.couldBeVerb_probability(word) and not word.endswith("ed"):
            return word, True
        result_list = self.convert(word, from_pos=self.WN_NOUN, to_pos=self.WN_VERB)
        if len(result_list) == 0:
            # parser这种单词形式
            if word.endswith("er"):
                t1 = word[0:-1]
                t2 = word[0:-2]
                if WordUtil.couldBeVerb(t1):
                    return t1, False
                if WordUtil.couldBeVerb(t2):
                    return t2, False
        else:
            if result_list[0][1] > 0.4:
                return result_list[0][0], False
        return None, False

    def noun_2_adj(self, word):
        """
        名词转adj
        :param word:
        :return:
        """
        result_list = self.convert(word, from_pos=self.WN_NOUN, to_pos=self.WN_ADJECTIVE)
        if len(result_list) != 0:
            if result_list[0][1] > 0.5:
                return result_list[0][0], False
        return None, False

    def verb_2_noun(self, word):
        """
        verb转noun
        :param word:
        :return:
        """
        result_list = self.convert(word, from_pos=self.WN_VERB, to_pos=self.WN_NOUN)
        if len(result_list) != 0:
            if result_list[0][1] > 0.5:
                return result_list[0][0], False
        return None, False

    def convert(self, word, from_pos, to_pos):
        """ Transform words given from/to POS tags """
        synsets = wn.synsets(word, pos=from_pos)
        if len(word) > 2:
            start_word = word[0:2]
        else:
            start_word = ""
        # Word not found
        if not synsets:
            return []
        # Get all lemmas of the word (consider 'a'and 's' equivalent)
        lemmas = []
        for s in synsets:
            for l in s.lemmas():
                if s.name().split('.')[1] == from_pos or from_pos in (
                        self.WN_ADJECTIVE, self.WN_ADJECTIVE_SATELLITE) and \
                        s.name().split('.')[1] in (self.WN_ADJECTIVE, self.WN_ADJECTIVE_SATELLITE):
                    lemmas += [l]

        # Get related forms
        derivationally_related_forms = [(l, l.derivationally_related_forms()) for l in lemmas]
        # TODO 过滤
        # filter only the desired pos (consider 'a' and 's' equivalent)
        related_noun_lemmas = []

        for drf in derivationally_related_forms:
            for l in drf[1]:
                if l.synset().name().split('.')[1] == to_pos or to_pos in (
                        self.WN_ADJECTIVE, self.WN_ADJECTIVE_SATELLITE) and \
                        l.synset().name().split('.')[1] in (self.WN_ADJECTIVE, self.WN_ADJECTIVE_SATELLITE):
                    if start_word != "" and len(l._name) > 2 and not str(l._name).startswith(start_word):
                        continue

                    related_noun_lemmas += [l]
        # Extract the words from the lemmas
        words = [l.name() for l in related_noun_lemmas]
        len_words = len(words)

        # Build the result in the form of a list containing tuples (word, probability)
        result = [(w, float(words.count(w)) / len_words) for w in set(words)]
        result.sort(key=lambda w: -w[1])

        # return all the possibilities sorted by probability
        return result


if __name__ == '__main__':
    vocabulary_conversion = VocabularyConversion()
    while True:
        t = input("input")
        a1 = vocabulary_conversion.convert(t, "n", "a")
        print(a1)
        a2 = vocabulary_conversion.convert(t, "v", "a")
        print(a2)
        a4 = vocabulary_conversion.convert(t, "r", "a")
        print(a4)
        # a = vocabulary_conversion.noun_2_verb('Field')

    a = vocabulary_conversion.convert('PutField', 'n', 'v')
    print(a)
#     # vocabulary_conversion.convert('writer', 'n', 'v')
#     # vocabulary_conversion.convert('quickly', 'r', 'a')
#     vocabulary_conversion.convert('parse', 'n', 'v')
#     a = vocabulary_conversion.noun_2_verb('writer')
#     print(a)
#     print("a")
